from sof3dcfun import *

def somf3d(dn,dipi,dipx,r1,r2,eps,order,option=1):
	'''
	somf3d: 3D structure-oriented median filter
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dipi: inline slope
	dipx: xline slope
	r1,r2:    spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	
	OUTPUT
	ds: filtered data
	 
	REFERENCES
	Huang et al., 2021, Erratic noise suppression using iterative structure-oriented space-varying median filtering with sparsity constraint, Geophysical Prospecting, 69, 101-121.
	Chen et al., 2020, Deblending of simultaneous-source data using a structure-oriented space-varying median filter, Geophysical Journal International, 222, 1805–1823.
	Gan et al., 2016, Separation of simultaneous sources using a structural-oriented median filter in the flattened dimension, Computers & Geosciences, 86, 46-54.
	Chen, Y., 2015, Deblending using a space-varying median filter, Exploration Geophysics, 46, 332-341.
	
	EXAMPLE
	demos/test_pyseistr_somf3d.py
	'''
	import numpy as np
	nnp=(2*r1+1)*(2*r2+1);
	
	#flattening
	from .pwspray3d import pwspray3d
	from .mf import mf
	from .mf import svmf
	
	u = pwspray3d(dn,dipi,dipx,r1,r2,order,eps);

	n3=dn.shape[2];
	n2=dn.shape[1];
	ns2=2*r1*r2+1;
	for i3 in range(0,n3):
		for i2 in range(0,n2):
			if option==1:
				u[:,:,i2,i3]=mf(u[:,:,i2,i3],ns2,1,2);
			else:
				u[:,:,i2,i3],win_tmp=svmf(u[:,:,i2,i3],ns2,1,2);
	
	ds=u[:,int(nnp/2),:,:];

	return ds


def somf3dc(dn,dipi,dipx,r1,r2,eps,order,option=1,verb=1):
	'''
	somf3dc: 3D structure-oriented median filter implemented in C
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dipi: inline slope
	dipx: xline slope
	r1,r2:    spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	
	OUTPUT
	ds: filtered data
	 
	REFERENCES
	Huang et al., 2021, Erratic noise suppression using iterative structure-oriented space-varying median filtering with sparsity constraint, Geophysical Prospecting, 69, 101-121.
	Chen et al., 2020, Deblending of simultaneous-source data using a structure-oriented space-varying median filter, Geophysical Journal International, 222, 1805–1823.
	Gan et al., 2016, Separation of simultaneous sources using a structural-oriented median filter in the flattened dimension, Computers & Geosciences, 86, 46-54.
	Chen, Y., 2015, Deblending using a space-varying median filter, Exploration Geophysics, 46, 332-341.
	
	EXAMPLE
	demos/test_pyseistr_somf3d.py
	'''
	import numpy as np
	nnp=(2*r1+1)*(2*r2+1);

	if dn.ndim==2:
		[n1,n2]=dn.shape;
		n3=1;
	else: #assuming ndim=3;
		[n1,n2,n3]=dn.shape;
		
	dn=np.float32(dn).flatten(order='F');
	dipi=np.float32(dipi).flatten(order='F');
	dipx=np.float32(dipx).flatten(order='F');
	
	rmf=2*r1*r2+1; #median filter length
	ds=csomf3d(dn,dipi,dipx,n1,n2,n3,r1,r2,rmf,option,order,eps,verb);
	ds=ds.reshape([n1,n2,n3],order='F');

	return ds




import numpy as np
def pwspray2d(u1,dip,nr,order,eps):
	'''
	pwspray2d: 2D plane-wave spray operator
	
	INPUT
	u1: noisy data  
	dip: slope 
	nr: spray radius
	order: PWD order
	eps: regularization (default:0.01);
	
	OUTPUT
	u: smooth data
   	'''
	[n1,n2]=u1.shape;
	ns=nr;	 	#spray radius
	ns2=2*ns+1;	#spray diameter

	n=n1*n2;nu=n1*n2*ns2;

	u=np.zeros(n1*ns2*n2);

	p=dip;
	trace=np.zeros(n1);
	e=eps*eps;
	nw=order;

	if n!=n1*n2:
		print('Wrong size 	%d != 	%d*	%d'%(n,n1,n2)); 


	if nu!=n*ns2:
		print('Wrong size 	%d != 	%d*	%d'%(nu,n,ns2)); 

	u1=u1.flatten(order='F')
	for i in range(0,n2):
		for i1 in range(0,n1):
			trace[i1]=u1[i*n1+i1];
			u[(i*ns2+ns)*n1+i1] = u[(i*ns2+ns)*n1+i1] + trace[i1];
	
		#predict forward
		#corresponding to the eq.24
		for iis in range(0,ns):
			ip=i-iis-1;
			if ip<0:
				break;
			j=ip*ns2+ns-iis-1;
			[w,diag,offd,trace] = predict_step(e,nw,0,p[:,ip],trace);
			for i1 in range(0,n1):
				u[j*n1+i1] = u[j*n1+i1] + trace[i1];
	
		for i1 in range(0,n1):
			trace[i1]=u1[i*n1+i1];

	
		# predict backward
		# 	#corresponding to the eq.25
		for iis in range(0,ns):
			ip=i+iis+1;
			if ip>=n2:
				break;
			j=ip*ns2+ns+iis+1;
			[w,diag,offd,trace] = predict_step(e,nw,1,p[:,ip-1],trace);
			for i1 in range(0,n1):
				u[j*n1+i1]= u[j*n1+i1]+trace[i1];

	return u

def predict_step(e,nw,forw,pp,trace1):
	'''
	predict_step: prediction step
	
	INPUT
	e: regularization parameter (default, 0.01*0.01);
	nw: accuracy order
	forw: forward or backward
	n1: trace length
	pp: slope
	trace: input trace
	
	OUTPUT
	w: PWD object
	diag,offd: diagonal/offdiagonals of the banded matrix
	trace: output trace
	'''
	n1=trace1.shape[0];
	nb=2*nw;
	eps=e;
	eps2=e;

	diag=np.zeros(n1);
	offd=np.zeros([n1,nb]);


	diag,offd = regularization(diag,offd,nw,eps,eps2);


	#only define diagonal,offdiagonal

	w,diag,offd = pwd_define(forw,diag,offd,n1,nw,pp);
   

	t0=trace1[0];
	t1=trace1[1];
	t2=trace1[n1-2];
	t3=trace1[n1-1];

	trace = pwd_set(0,w,diag,offd,pp,trace1);


	trace[0]=trace[0]+eps2*t0;
	trace[1]=trace[1]+eps2*t1;

	trace[n1-2]=trace[n1-2]+eps2*t2;
	trace[n1-1]=trace[n1-1]+eps2*t3;


	trace = banded_solve(n1,nb,diag,offd,trace);   

	return w,diag,offd,trace


def regularization(diag,offd,nw,eps,eps2):
	'''
	fill diag and offd using regularization 
	
	INPUT
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	eps: regularization parameter (default: e*e, e=0.01);
	eps2: second regularization parameter (default, same as eps)
	nw: accuracy order (nb=2*nw)
	
	OUTPUT
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	#
	'''

	nb=2*nw;
	n1=diag.size;

	for i1 in range(0,n1):
		diag[i1]=6*eps;
		offd[i1,0]=-4*eps;
		offd[i1,1]=eps;
	
		for ib in range(2,nb):
			offd[i1,ib]=0.0;

		diag[0]=eps2+eps;
		diag[1]=eps2+5*eps;
	
		diag[n1-1]=eps2+eps;
		diag[n1-2]=eps2+5*eps;  
	
		offd[0,0]=-2*eps;
		offd[n1-2,0]=-2*eps;

	return diag,offd

def pwd_define(forw,diag,offd,n1,nw,pp):
	'''
	pwd_define: matrix multiplication
	
	INPUT
	forw:forward or backward
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	n1: trace length
	nw: PWD filter(accuracy) order (default nw=1)
	pp: slope				  (1D array)
	
	OUTPUT
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	w: PWD object(struct)
	
	'''
	
	#define PWD object(struct)
	w={'n':n1,'na':2*nw+1,'a':np.zeros([n1,2*nw+1]),'b':np.zeros(2*nw+1)}
	
	n=w['n'];
	nb=2*nw;

	for i in range(0,n):
		w['b']=passfilter(pp[i],nw)[0];
		for j in range(0,w['na']):
			if forw:
				w['a'][i,j]=w['b'][w['na']-j-1];
			else:
				w['a'][i,j]=w['b'][j];

	for i in range(0,n):
		for j in range(0,w['na']):
			k=i+j-nw;
			if k>=nw and k<n-nw:
				aj=w['a'][k,j];
				diag[i]=diag[i]+aj*aj;
		for m in range(0,2*nw):
			for j in range(m+1,w['na']):
				k=i+j-nw;
				if k>=nw and k<n-nw:
					aj=w['a'][k,j];
					am=w['a'][k,j-m-1];
					offd[i,m]=offd[i,m]+am*aj;
	return w,diag,offd


def passfilter(p,nw):
	'''
	passfilter: find filter coefficients 
	All-pass plane-wave destruction filter coefficients
	
	INPUT
	p: slope
	nw: PWD filter(accuracy) order
	
	OUTPUT
	a: output filter (n+1) (1D array)
	b: temp variable of a
	
	'''
	n=nw*2;
	b=np.zeros(n+1);
	a=np.zeros(n+1);
	for k in range(0,n+1):
		bk=1;
		for j in range(0,n):
			if j<n-k:
				bk=bk*(k+j+1.0)/(2*(2*j+1)*(j+1));
			else:
				bk=bk*1.0/(2*(2*j+1));
		b[k]=bk;
	for k in range(0,n+1):
		ak=b[k];
		for j in range(0,n):
			if j<n-k:
				ak=ak*(n-j-p);
			else:
				ak=ak*(p+j+1);
		a[k]=ak;
	return a,b

def pwd_set(adj,w,diag,offd,pp,inp):
	'''
	pwd_set: matrix multiplication
	
	INPUT
	adj:adjoint flag
	w: PWD object(struct)
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	pp: slope				  (1D array)
	inp: model
	
	OUTPUT
	out: data 
	
	'''

	n=w['n'];
	nw=np.int((w['na']-1)/2);

	#	pwd_set
	tmp=np.zeros(n);
	out=np.zeros(n);
	if adj:
		for i in range(0,n):
			tmp[i]=0.0;

		for i in range(0,n):
			for j in range(0,w['na']):
				k=i+j-nw;
				if k>=nw and k<n-nw:
					tmp[k]=tmp[k]+w['a'][k,j]*inp[i];
		for i in range(0,n):
			out[i]=0.0;
		for i in range(nw,n-nw):
			for j in range(0,w['na']):
				k=i+j-nw;
				out[k]=out[k]+w['a'][i,j]*tmp[i];
	else:
		for i in range(0,n):
			tmp[i]=0.0;

		for i in range(nw,n-nw):
			for j in range(0,w['na']):
				k=i+j-nw;
				tmp[i]=tmp[i]+w['a'][i,j]*inp[k];
		for i in range(0,n):		
			out[i]=0.0;
			for j in range(0,w['na']):
				k=i+j-nw;
				if k>=nw and k<n-nw:
					out[i]=out[i]+w['a'][k,j]*tmp[k];
	return out

def banded_solve(n,band,diag,offd,b):
	'''
	banded_solve: Banded matrix solver
	
	INPUT
	n:	matrix size
	band: band size
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	b: input trace
	
	OUTPUT
	b: trace solution
	
	'''

	#define Band object(struct)
	slv={'n':n,'band':band,'d':np.zeros(n),'o':np.zeros([n-1,band])}

	# define the banded matrix
	for k in range(0,slv['n']):
		t=diag[k];
		m1=np.min([k,slv['band']]);
		for m in range(0,m1):
			t=t-slv['o'][k-m-1,m]*slv['o'][k-m-1,m]*slv['d'][k-m-1];

		slv['d'][k]=t;
		n1=np.min([slv['n']-k-1,slv['band']]);
		for n in range(0,n1):
			t=offd[k,n];
			m1=np.min([k,slv['band']-n-1]);
			for m in range(0,m1):
				t=t-slv['o'][k-m-1,m]*slv['o'][k-m-1,n+m+1]*slv['d'][k-m-1];

			slv['o'][k,n]=t/slv['d'][k];

	# the solver 
	for k in range(1,slv['n']):
		t=b[k];
		m1=np.min([k,slv['band']]);
		for m in range(0,m1):
			t=t-slv['o'][k-m-1,m]*b[k-m-1];
		b[k]=t;

	for k in range(slv['n']-1,-1,-1):
		t=b[k]/slv['d'][k];
		m1=np.min([slv['n']-k-1,slv['band']]);
		for m in range(0,m1):
			t=t-slv['o'][k,m]*b[k+m+1];

		b[k]=t;


	return b


def ricker(f,dt,tlength=None):
	'''
	ricker: Ricker wavelet of central frequency f.
	
	INPUT:
	f : central freq. in Hz (f <<1/(2dt) )
	dt: sampling interval in sec
	tlength : the duration of wavelet in sec
	
	OUTPUT: 
	w:  the Ricker wavelet
	tw: time axis
	
	EXAMPLE
	from pyseistr import ricker;
	wav,tw=ricker(20,0.004,2);
	import matplotlib.pyplot as plt;
	plt.plot(tw,wav);plt.xlabel('Time (s)');plt.ylabel('Amplitude');plt.show();
	
	'''
	import numpy as np
	
	if tlength!=None:
		nw=np.floor(tlength/dt)+1;
	else:
		nw=2.2/f/dt;
		nw=2*np.floor(nw/2)+1;
	nc=np.floor(nw/2);
	nw=int(nw)
	w =np.zeros(nw);
	
	k=np.arange(1,nw+1,1);
	alpha = (nc-k+1)*f*dt*np.pi;
	beta=np.power(alpha,2);
	w = (1.-beta*2)*np.exp(-beta);
	tw = -(nc+1-k)*dt;
	return w,tw


def mf(D,nfw=7,ifb=1,axis=2):
	'''
	MF: median filter along first or second axis for 2D profile
	
	INPUT 
	D:   	intput data 
	nfw:    window size 
	ifb:    if use padded boundary (if not, zero will be padded)
	axis:   along the vertical (1) or horizontal (2) axis
	    
	OUTPUT 
	D1:  	output data
	
	Copyright (C) 2014 The University of Texas at Austin
	Copyright (C) 2014 Yangkang Chen
	Ported to python in Apr, 17, 2022
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published
	by the Free Software Foundation, either version 3 of the License, or
	any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details: http://www.gnu.org/licenses/
	
	References
	Huang et al., 2021, Erratic noise suppression using iterative structure-oriented space-varying median filtering with sparsity constraint, Geophysical Prospecting, 69, 101-121.
	Chen et al., 2020, Deblending of simultaneous-source data using a structure-oriented space-varying median filter, Geophysical Journal International, 222, 1805–1823.
	Gan et al., 2016, Separation of simultaneous sources using a structural-oriented median filter in the flattened dimension, Computers & Geosciences, 86, 46-54.
	Chen, Y., 2015, Deblending using a space-varying median filter, Exploration Geophysics, 46, 332-341.
	'''
	import numpy as np

	# nfw should be odd
	if np.mod(nfw,2)==0:
		nfw=nfw+1;

	if axis==2:
		D=D.transpose();
	n1=D.shape[0];
	n2=D.shape[1];
	
	nfw2=(nfw-1)/2;nfw2=int(nfw2);
	
	if ifb==1:
		D=np.concatenate((np.flipud(D[0:nfw2,:]),D,np.flipud(D[n1-nfw2:n1,:])),axis=0);
	else: 
		D=np.concatenate((np.zeros([nfw2,n2]),D,np.zeros([nfw2,n2])),axis=0);
	# output data
	D1=np.zeros([n1,n2]);
	for i2 in range(0,n2):
		for i1 in range(0,n1):
			D1[i1,i2]=np.median(D[i1:i1+nfw,i2]); 
	if axis==2:
		D1=D1.transpose();
		
	return D1


def svmf(D,nfw=7,ifb=1,axis=2,l1=2,l2=0,l3=2,l4=4):
	'''
	SVMF: space-varying median filter along first or second axis for 2D profile
	
	INPUT   
	D:   	intput data 
	nfw:    window size
	ifb:    if use padded boundary (if not, zero will be padded)
	axis:   along the vertical (1) or horizontal (2) axis
	    
	OUTPUT  
	D1:  	output data
	# 		 win_len: window length distribution
	
	Copyright (C) 2019 Yangkang Chen
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published
	by the Free Software Foundation, either version 3 of the License, or
	any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details: http://www.gnu.org/licenses/
	'''
	import numpy as np
	n1=D.shape[0];
	n2=D.shape[1];
	
	
	Dtmp=mf(D,nfw,ifb,axis);
	medianv=np.sum(np.abs(Dtmp.flatten()))/(n1*n2);

	# nfw should be odd
	if np.mod(nfw,2)==0:
		nfw=nfw+1;

	# calculate length
	win_len=np.zeros([n1,n2],dtype='int');
	for i2 in range(0,n2):
		for i1 in range(0,n1):
			if np.abs(Dtmp[i1,i2]) < medianv:
				if np.abs(Dtmp[i1,i2]) < medianv/2:
					win_len[i1,i2]=nfw+l1;
				else:
					win_len[i1,i2]=nfw+l2;
			else:
				if np.abs(Dtmp[i1,i2]) > medianv*2:
					win_len[i1,i2]=nfw-l4;
				else:
					win_len[i1,i2]=nfw-l3;
	if axis==2:
		D=D.transpose();
		win_len=win_len.transpose();
	n1=D.shape[0];
	n2=D.shape[1];
	win_len2=(win_len-1)/2;win_len2=win_len2.astype(int);

	nfw_b=(np.max([nfw+l1,nfw+l2])-1)/2;nfw_b=int(nfw_b);
	
	if ifb==1:
		D=np.concatenate((np.flipud(D[0:nfw_b,:]),D,np.flipud(D[n1-nfw_b:n1,:])),axis=0);
	else:
		D=np.concatenate((np.zeros([nfw_b,n2]),D,np.zeros([nfw_b,n2])),axis=0);
	
	# output data
	D1=np.zeros([n1,n2]);
	for i2 in range(0,n2):
		for i1 in range(0,n1):
# 			print(nfw_b,win_len2[i1,i2],i1+nfw_b-win_len2[i1,i2],i1+nfw_b+win_len2[i1,i2])
# 			print(D.shape)
# 			print(i1,i2,n1,n2)
			D1[i1,i2]=np.median(D[i1+nfw_b-win_len2[i1,i2]:i1+nfw_b+win_len2[i1,i2]+1,i2]); 
	win_len=win_len2*2+1;
	
	if axis==2:
		D1=D1.transpose();
		win_len=win_len.transpose();    

	return D1,win_len



import numpy as np
from cohcfun import *

def coh(inpt,par,operator=-1):
    '''
    Coherency measure based on the 
    Forward and Adjoint operators for Radon RT in the time domain.
    IN   	in:   		intput data 
      	Param:  		parameters combination
    par.h:		offset
    par.v:		velocity
    par.nt:		number of samples
    par.dt:		time interval
    par.typ:    1: linear 2: parablic 3: hyperbolic

   	operator: 
# 			operator =  1 means impute is m(tau,v) and output is d(t,x) FORWARD  OP
#			operator = -1 means input is d(t,x) and output is m(tau,v)  ADJOINT  OP 
     
 	OUT   out:  		output data

	 Copyright (C) 2014 The University of Texas at Austin
 	Copyright (C) 2014 Yangkang Chen

	REFERENCE
	Chen, 2018, GEO, Automatic velocity analysis using high-resolution hyperbolic Radon transform
	Chen et al., 2023, GEO, Enhancing earthquake detection from distributed acoustic sensing data by coherency measure and moving-rank-reduction filtering
    '''
    
    h = par['h'];
    v = par['v'];
    nt = par['nt'];
    dt = par['dt'];
    typ = par['typ'];
    
    nh = len(h)
    nv = len(v)
    
    if operator == -1:
        m = np.zeros((nt,nv),dtype='float32')
        num=np.zeros((nt,nv),dtype='float32')
        den=np.zeros((nt,nv),dtype='float32')
    if operator ==  1:
        d = np.zeros((nt,nh),dtype='float32')
    
    if operator == -1:
        d = inpt
    if operator ==  1:
        m = inpt
    
    hmax=np.max(abs(h))
    
    for itau in range(nt):
        #print(itau)
        for ih in range(nh):
            for iv in range(nv):
## This can also be replaced by Parabolic or linear integration 
                if typ == 1:
                    t = itau*dt + h[ih]*v[iv]	
                    it = int(np.floor(t/dt)+1)     
                elif typ == 2:
                    t = itau*dt + h[ih]*h[ih]*v[iv]/hmax/hmax  #curvature
                    it = int(np.floor(t/dt)+1)
            #if(it<=0) it=1;end
                elif typ == 3:
                    t = np.sqrt ((itau*dt)^2 + (h[ih]/v[iv])^2 )
                    it = int(np.floor(t/dt)+1)
                else:
                    t = np.sqrt ((itau*dt)^2 + (h[ih]/v[iv])^2 )
                    it = int(np.floor(t/dt)+1)
                #print(it),print(ih),print(itau),print(iv,'\n')
                if (it+1<=nt) & (it+1>0):  
                    if (operator==-1):
#                         m[itau,iv]=m[itau,iv]+d[it,ih]
                        num[itau,iv]=num[itau,iv]+d[it,ih]
                        den[itau,iv]=den[itau,iv]+d[it,ih]*d[it,ih]
                        
                    if (operator==1):
                        d[it,ih]=d[it,ih]+m[itau,iv]
    
    m=(num*num)/(nh*den);
    
    
    if operator == 1:
        out = d
    if operator ==-1:
        out = m

    return out
    
    
def cohc(din,par,operator=-1,verb=0):
	'''
	Forward and Adjoint operators for Radon RT in the time domain.
	radon implemented in C
	'''
	
	h = par['h'];
	v = par['v'];
	nt = par['nt'];
	dt = par['dt'];
	typ = par['typ'];

	nh = len(h);
	nv = len(v);
	
	din=np.float32(din.flatten(order='F'));
	v=np.float32(v);
	h=np.float32(h);
	
	dout=cohc_fb(din, v, h, typ, nt, nv, nh, dt, operator, verb);
	
	if operator==1: #forward
		dout=dout.reshape(nt,nh,order='F');
	if operator==-1: #backward/adjoint
		dout=dout.reshape(nt,nv,order='F');
	
	
	return dout
	
	



from matplotlib.colors import ListedColormap
import numpy as np

def cseis():
	seis=np.concatenate(
(np.concatenate((0.5*np.ones([1,40]),np.expand_dims(np.linspace(0.5,1,88),axis=1).transpose(),np.expand_dims(np.linspace(1,0,88),axis=1).transpose(),np.zeros([1,40])),axis=1).transpose(),
np.concatenate((0.25*np.ones([1,40]),np.expand_dims(np.linspace(0.25,1,88),axis=1).transpose(),np.expand_dims(np.linspace(1,0,88),axis=1).transpose(),np.zeros([1,40])),axis=1).transpose(),
np.concatenate((np.zeros([1,40]),np.expand_dims(np.linspace(0,1,88),axis=1).transpose(),np.expand_dims(np.linspace(1,0,88),axis=1).transpose(),np.zeros([1,40])),axis=1).transpose()),axis=1)

	return ListedColormap(seis)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    Yangkang Chen (chenyk2016@gmail.com), 2021-2023   
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

# import collections
import logging

class PyseistrError(Exception):
    """
    Base class for all Pyseistr exceptions. Will probably be used for all
    exceptions to not overcomplicate things as the whole package is pretty
    small.
    """
    pass


class PyseistrWarning(UserWarning):
    """
    Base class for all Pyseistr warnings.
    """
    pass

__version__ = "0.1.0"

# Setup the logger.
logger = logging.getLogger("Pyseistr")
logger.setLevel(logging.WARNING)
# Prevent propagating to higher loggers.
logger.propagate = 0
# Console log handler.
ch = logging.StreamHandler()
# Add formatter
FORMAT = "[%(asctime)s] - %(name)s - %(levelname)s: %(message)s"
formatter = logging.Formatter(FORMAT)
ch.setFormatter(formatter)
logger.addHandler(ch)

from .dip2d import dip2d
from .dip3d import dip3d
from .divne import divne
from .somean2d import somean2d
from .somean3d import somean3d
from .somf2d import somf2d
from .somf3d import somf3d
from .soint2d import soint2d #structure-oriented interpolation for 2D data
from .soint3d import soint3d #structure-oriented interpolation for 3D data
from .ricker import ricker
from .bp import bandpass
from .fk import fkdip
from .plot import cseis

from .das import coh 
from .das import cohc

from .synthetics import *

## C versions
from .dip2d import dip2dc
from .dip3d import dip3dc
from .somean2d import somean2dc
from .somean3d import somean3dc
from .somf2d import somf2dc
from .somf3d import somf3dc
from .soint2d import soint2dc
from .soint3d import soint3dc
from .bp import bandpassc

from dipcfun import *
from sofcfun import *
from sof3dcfun import *
from bpcfun import *








from sofcfun import *
from sof3dcfun import *

def somean3d(dn,dipi,dipx,r1,r2,eps,order):
	'''
	somean3d: 3D structure-oriented mean filter
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dipi: inline slope
	dipx: xline slope
	r1,r2:    spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	
	OUTPUT
	ds: smoothed data
	
	EXAMPLE
	
	'''
	import numpy as np
	nnp=(2*r1+1)*(2*r2+1);
	
	#flattening
	from .pwspray3d import pwspray3d
	u = pwspray3d(dn,dipi,dipx,r1,r2,order,eps);

	#smoothing
	ds=np.sum(u,1)/nnp;

	return ds

def somean3dc(dn,dipi,dipx,r1,r2,eps,order,verb=0):
	'''
	somean3d: 3D structure-oriented mean filter
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dipi: inline slope
	dipx: xline slope
	r1,r2:    spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	
	OUTPUT
	ds: smoothed data
	
	EXAMPLE
	
	'''
	import numpy as np

	if dn.ndim==2:
		[n1,n2]=dn.shape;
		n3=1;
	else: #assuming ndim=3;
		[n1,n2,n3]=dn.shape;
		
	dn=np.float32(dn).flatten(order='F');
	dipi=np.float32(dipi).flatten(order='F');
	dipx=np.float32(dipx).flatten(order='F');
	
# 	print(dn.max(),dn.min(),dn.var())
	ds=csomean3d(dn,dipi,dipx,n1,n2,n3,r1,r2,order,eps,verb);
	ds=ds.reshape([n1,n2,n3],order='F');
# 	print(ds.max(),ds.min(),ds.var())
	
	return ds





from soint3dcfun import *

def soint3d(din,mask,dipi,dipx,order=1,niter=100,njs=[1,1],drift=0,verb=1):
	'''
	soint3d: 3D structure-oriented interpolation
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dipi: inline slope
	dipx: xline slope
	order:    PWD order
	
	OUTPUT
	dout: filtered data 

	EXAMPLE
	demos/test_pyseistr_soint3d.py
	'''
	from .solvers import solver
	from .solvers import cgstep
	from .operators import allpass3_lop
	import numpy as np
	
	nw=order;
	nj1=njs[0];
	nj2=njs[1];

	[n1,n2,n3]=din.shape;
	n123=n1*n2*n3;
	
	mm=din;
	mm=mm.flatten(order='F');
	known=np.zeros([n123,1]);
	
	if mask  is not None:
		dd=mask.flatten(order='F');
		for ii in range(0,n123):
			known[ii] = (dd[ii] !=0) ;
			dd[ii]=0;
	else:
		for ii in range(0,n123):
			known[ii] = (mm[ii] !=0) ;
			dd[ii]=0;

	pp=dipi.flatten(order='F');
	qq=dipx.flatten(order='F');
	
	ap1={'nw':nw,'nj':nj1,'nx':n1,'ny':n2,'nz':n3,'drift':drift,'pp':pp};
	ap2={'nw':nw,'nj':nj2,'nx':n1,'ny':n2,'nz':n3,'drift':drift,'pp':qq};
	
	par_L={'ap1':ap1,'ap2':ap2,'nm':n123,'nd':2*n123};
	
	par_sol={'known':known,'x0':mm,'verb':1};
	
	dd=np.zeros([2*n123,1]);
	mm2,tmp=solver(allpass3_lop,cgstep,n123,2*n123,mm,dd,niter,par_L,par_sol);
	
	dout=mm2.reshape(n1,n2,n3,order='F');

	return dout
	
	
def soint3dc(din,mask,dipi,dipx,order=1,niter=100,njs=[1,1],drift=0,seed=202223,hasmask=1,var=0,verb=1):
	'''
	soint3dc: 3D structure-oriented interpolation implemented in C
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dipi: inline slope
	dipx: xline slope
	r1,r2:    spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	hasmask: if 1, using the provided mask; if 0, using the data itself to determine
	
	OUTPUT
	dout: filtered data 
	
	EXAMPLE
	demos/test_pyseistr_soint3d.py
	'''
	
	from .solvers import solver
	from .solvers import cgstep
	from .operators import allpass3_lop
	import numpy as np
	
	nw=order;
	nj1=njs[0];
	nj2=njs[1];
	[n1,n2,n3]=din.shape;

	
	dipi=np.float32(dipi).flatten(order='F');
	dipx=np.float32(dipx).flatten(order='F');
	mask=np.float32(mask).flatten(order='F');
	din=np.float32(din).flatten(order='F');
	dout=csoint3d(din,mask,dipi,dipx,n1,n2,n3,nw,nj1,nj2,niter,drift,seed,hasmask,var,verb);
	dout=dout.reshape(n1,n2,n3,order='F');


	
	
	return dout






def gensyn():
	'''
	quickgen: quickly generate the representative synthetic data used in the paper
	
	INPUT
	None
	
	OUTPUT
	data: synthetic data
	
	EXAMPLE
	from pyseistr import gensyn
	data=gensyn();
	import matplotlib.pyplot as plt;
	plt.imshow(data);plt.ylabel('Time sample');plt.xlabel('Trace');plt.show();

	'''
	import numpy as np
	from .ricker import ricker
	[w,tw]=ricker(30,0.001,0.1);
	t=np.zeros([300,1000]);sigma=300;A=100;B=200;
	data=np.zeros([400,1000]);

	[m,n]=t.shape;
	for i in range(1,n+1):
		k=np.floor(-A*np.exp(-np.power(i-n/2,2)/np.power(sigma,2))+B);k=int(k);
		if k>1 and k<=m:
			t[k-1,i-1]=1;

	for i in range(1,n+1):
		data[:,i-1]=np.convolve(t[:,i-1],w);
	
	return data
	
	
def genflat(nt=100,nx=60,t=[30,50,80],amp=[1,1,1],freq=30,dt=0.004):
	'''
	quickgen: quickly generate the representative synthetic data used in the paper
	
	INPUT
	None
	
	OUTPUT
	data: synthetic data
	
	EXAMPLE
	from pyseistr import genflat
	data=genflat();
	import matplotlib.pyplot as plt;
	plt.imshow(data);plt.ylabel('Time sample');plt.xlabel('Trace');plt.show();

	'''
	import numpy as np
	from .ricker import ricker
	[w,tw]=ricker(freq,dt,60*dt);
	data=np.zeros([nt,nx]);

	trace=np.zeros(nt)
	for ii in range(len(t)):
		trace[t[ii]]=amp[ii];

	for ii in range(nx):
		data[:,ii]=np.convolve(trace,w,mode='same');
	
	return data


import numpy as np
from dipcfun import *

def dip3d(din,niter=5,liter=10,order=2,eps_dv=0.01, eps_cg=1, tol_cg=0.000001,rect=[5,5,5],verb=1):
	'''
	dip3d: 3D dip estimation based on shaping regularized PWD algorithm
	(independent implementation)
	
	Ported to Python by Yangkang Chen, 2022, verified to be exactly the same as the Matlab version
	
	INPUT
	din: input data (nt*nx)
	niter: number of nonlinear iterations
	liter: number of linear iterations (in divn)
	order: accuracy order
	eps_dv: eps for divn  (default: 0.01)
	eps_cg: eps for CG	(default: 1)
	tol_cg: tolerence for CG (default: 0.000001)
	rect:  smoothing radius (ndim*1)
	verb: verbosity flag
	
	OUTPUT
	dipi:  inline 3D slope
	dipx:  xline 3D slope
	
	NOTE: mask function is not enabled yet (will do in the future)
	'''
	from .divne import divne
	
	dim = 3;
	n = np.zeros(dim,dtype='int');
	n1 = din.shape[0];
	n2 = din.shape[1];
	n3 = din.shape[2];
	n[0] = n1;
	n[1] = n2;
	n[2] = n3;

	n123 = din.size;

	dip_i=np.zeros([n1,n2,n3]);
	dip_x=np.zeros([n1,n2,n3]);

	for iter in range(0,niter):

		#corresponding to the eq.21 in the paper
		u1_i,u2_i = conv_allpass_i(din,dip_i,order); 	#inline linearization using the updated dip	
		ratio_i  = divne(-u2_i, u1_i, liter, rect, n, eps_dv, eps_cg, tol_cg,verb);

		#corresponding to the eq.21 in the paper
		u1_x,u2_x = conv_allpass_x(din,dip_x,order); 	#xline linearization using the updated dip
		ratio_x  = divne(-u2_x, u1_x, liter, rect, n, eps_dv, eps_cg, tol_cg,verb);
	
		dip_i=dip_i+ratio_i;
		dip_x=dip_x+ratio_x;

	return dip_i,dip_x

def dip3dc(din,niter=5,liter=10,order=2,eps_dv=0.01, eps_cg=1, tol_cg=0.000001,rect=[5,5,5],verb=1,runc=1,mask=None):
	'''
	dip3dc: 3D dip estimation based on shaping regularized PWD algorithm
	(C implementation)
	
	Ported to Python by Yangkang Chen, 2022, verified to be exactly the same as the Matlab version
	
	INPUT
	din: input data (nt*nx)
	niter: number of nonlinear iterations
	liter: number of linear iterations (in divn)
	order: accuracy order
	eps_dv: eps for divn  (default: 0.01)
	eps_cg: eps for CG	(default: 1)
	tol_cg: tolerence for CG (default: 0.000001)
	rect:  smoothing radius (ndim*1)
	verb: verbosity flag
	runc: if runc
	 
	OUTPUT
	dipi:  inline 3D slope
	dipx:  xline 3D slope
	'''
	from .divne import divne
	
	dim = 3;
	n = np.zeros(dim,dtype='int');
	n1 = din.shape[0];
	n2 = din.shape[1];
	n3 = din.shape[2];
	n[0] = n1;
	n[1] = n2;
	n[2] = n3;

	n123 = din.size;

	r1=rect[0]
	r2=rect[1]
	r3=rect[2]
	
	if mask is None:
		din=np.float32(din.flatten(order='F'));
		hasmask=0
	else:
		if din.size != mask.size:
			Exception("Mask and Data should have the same dimension")
		else:
			din=np.float32(np.concatenate([din.flatten(order='F'),mask.flatten(order='F')],axis=0));
			hasmask=1;
	
	dip=dipc(din,n1,n2,n3,niter,liter,order,eps_dv,eps_cg,tol_cg,r1,r2,r3,hasmask,verb);
	dip=dip.reshape(n1,n2,n3,2,order='F');
	
	dip_i=dip[:,:,:,0]
	dip_x=dip[:,:,:,1]	
	

	return dip_i,dip_x


def conv_allpass_i(din,dip,order):
	'''
	conv_allpass_i: Convolutional operator implemented by an allpass filter (iline direction)
	
	Linearized inverse problem
	C'(\sigma)d\Delta \sigma =  C(\sigma)d
	
	IPNUT:
	din: input data
	dip: 3D dip
	order: accuracy order
	
	OUTPUT:
	u1: C'(\sigma)d (denominator)
	u2: C(\sigma)d  (numerator)
	
	'''
	[n1,n2,n3]=din.shape;
	u1=np.zeros([n1,n2,n3]);
	u2=np.zeros([n1,n2,n3]);

	if order==1:
		nw=1;
	else:
		nw=2;


	filt1=np.zeros(2*nw+1);
	filt2=np.zeros(2*nw+1);

	# cprresponding to eqs.17-19
	# destruct the data
	for i1 in range(nw,n1-nw):
		for i2 in range(0,n2-1):
			for i3 in range(0,n3):
				if order==1:
					filt1=B3d(dip[i1,i2,i3]);
					filt2=B3(dip[i1,i2,i3]);
				else:
					filt1=B5d(dip[i1,i2,i3]);
					filt2=B5(dip[i1,i2,i3]);

				for iw in range(-nw,nw+1):
					u1[i1,i2,i3]=u1[i1,i2,i3]+(din[i1+iw,i2+1,i3]-din[i1-iw,i2,i3])*filt1[iw+nw];
					u2[i1,i2,i3]=u2[i1,i2,i3]+(din[i1+iw,i2+1,i3]-din[i1-iw,i2,i3])*filt2[iw+nw];

	return u1,u2

def conv_allpass_x(din,dip,order):
	'''
	conv_allpass_x: Convolutional operator implemented by an allpass filter (xline direction)
	
	Linearized inverse problem
	C'(\sigma)d\Delta \sigma =  C(\sigma)d
	
	IPNUT:
	din: input data
	dip: 3D dip
	order: accuracy order
	
	OUTPUT:
	u1: C'(\sigma)d (denominator)
	u2: C(\sigma)d  (numerator)
	
	'''
	
	[n1,n2,n3]=din.shape;
	u1=np.zeros([n1,n2,n3]);
	u2=np.zeros([n1,n2,n3]);

	if order==1:
		nw=1;
	else:
		nw=2;


	filt1=np.zeros(2*nw+1);
	filt2=np.zeros(2*nw+1);

	# cprresponding to eqs.17-19
	# destruct the data
	for i1 in range(nw,n1-nw):
		for i2 in range(0,n2):
			for i3 in range(0,n3-1):
				if order==1:
					filt1=B3d(dip[i1,i2,i3]);
					filt2=B3(dip[i1,i2,i3]);
				else:
					filt1=B5d(dip[i1,i2,i3]);
					filt2=B5(dip[i1,i2,i3]);
				for iw in range(-nw,nw+1):
					u1[i1,i2,i3]=u1[i1,i2,i3]+(din[i1+iw,i2,i3+1]-din[i1-iw,i2,i3])*filt1[iw+nw];
					u2[i1,i2,i3]=u2[i1,i2,i3]+(din[i1+iw,i2,i3+1]-din[i1-iw,i2,i3])*filt2[iw+nw];
	return u1,u2


#cprresponding to eqs.13-16
#form the filters
def B3(sigma):
	'''
	B3 coefficient
	sigma: slope
	'''
	b3=np.zeros[2];
	b3[0]=(1-sigma)*(2-sigma)/12;
	b3[1]=(2+sigma)*(2-sigma)/6;
	b3[2]=(1+sigma)*(2+sigma)/12;
	
	return b3

def B3d(sigma):
	'''
	B3 coefficient derivative
	sigma: slope
	'''
	b3d=np.zeros[2];
	b3d[0]=-(2-sigma)/12-(1-sigma)/12;
	b3d[1]=(2-sigma)/6-(2+sigma)/6;
	b3d[2]=(2+sigma)/12+(1+sigma)/12;

	return b3d

def B5(sigma):
	'''
	B5 coefficient
	sigma: slope
	'''
	b5=np.zeros(5);
	b5[0]=(1-sigma)*(2-sigma)*(3-sigma)*(4-sigma)/1680;
	b5[1]=(4-sigma)*(2-sigma)*(3-sigma)*(4+sigma)/420;
	b5[2]=(4-sigma)*(3-sigma)*(3+sigma)*(4+sigma)/280;
	b5[3]=(4-sigma)*(2+sigma)*(3+sigma)*(4+sigma)/420;
	b5[4]=(1+sigma)*(2+sigma)*(3+sigma)*(4+sigma)/1680;

	return b5

def B5d(sigma):
	'''
	B5 coefficient derivative
	sigma: slope
	'''
	b5d=np.zeros(5);
	b5d[0]=-(2-sigma)*(3-sigma)*(4-sigma)/1680-\
	(1-sigma)*(3-sigma)*(4-sigma)/1680-\
	(1-sigma)*(2-sigma)*(4-sigma)/1680-\
	(1-sigma)*(2-sigma)*(3-sigma)/1680;

	b5d[1]=-(2-sigma)*(3-sigma)*(4+sigma)/420-\
	(4-sigma)*(3-sigma)*(4+sigma)/420-\
	(4-sigma)*(2-sigma)*(4+sigma)/420+\
	(4-sigma)*(2-sigma)*(3-sigma)/420;

	b5d[2]=-(3-sigma)*(3+sigma)*(4+sigma)/280-\
	(4-sigma)*(3+sigma)*(4+sigma)/280+\
	(4-sigma)*(3-sigma)*(4+sigma)/280+\
	(4-sigma)*(3-sigma)*(3+sigma)/280;

	b5d[3]=-(2+sigma)*(3+sigma)*(4+sigma)/420+\
	(4-sigma)*(3+sigma)*(4+sigma)/420+\
	(4-sigma)*(2+sigma)*(4+sigma)/420+\
	(4-sigma)*(2+sigma)*(3+sigma)/420;

	b5d[4]=(2+sigma)*(3+sigma)*(4+sigma)/1680+\
	(1+sigma)*(3+sigma)*(4+sigma)/1680+\
	(1+sigma)*(2+sigma)*(4+sigma)/1680+\
	(1+sigma)*(2+sigma)*(3+sigma)/1680;

	return b5d
	





from soint2dcfun import *

def soint2d(din,mask,dip,order=1,niter=100,njs=[1,1],drift=0,verb=1):
	'''
	soint2d: 2D structure-oriented interpolation
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dip: slope
	order:    PWD order
	eps: regularization (default:0.01);
	
	OUTPUT
	ds: filtered data 
	
	EXAMPLE
	demos/test_pyseistr_soint2d.py
	'''
	from .solvers import solver
	from .solvers import cgstep
	from .operators import allpass3_lop
	import numpy as np
	
	nw=order;
	nj1=njs[0];
	nj2=njs[1];

	[n1,n2]=din.shape;
	n12=n1*n2;
	
	mm=din;
	mm=mm.flatten(order='F');
	known=np.zeros([n12,1]);
	
	if mask  is not None:
		dd=mask.flatten(order='F');
		for ii in range(0,n12):
			known[ii] = (dd[ii] !=0) ;
			dd[ii]=0;
	else:
		for ii in range(0,n12):
			known[ii] = (mm[ii] !=0) ;
			dd[ii]=0;

	pp=dipi.flatten(order='F');
	
	dout=din

	return dout
	
	
def soint2dc(din,mask,dip,order=1,niter=100,njs=[1,1],drift=0,hasmask=1,twoplane=0,prec=0,verb=1):
	'''
	soint2d: 3D structure-oriented interpolation
	
	by Yangkang Chen, 2022
	
	INPUT
	dn: model  noisy data
	dipi: inline slope
	dipx: xline slope
	r1,r2:    spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	hasmask: if 1, using the provided mask; if 0, using the data itself to determine
	
	OUTPUT
	ds: filtered data 
	
	EXAMPLE
	demos/test_pyseistr_soint3d.py
	'''
	from .solvers import solver
	from .solvers import cgstep
	from .operators import allpass3_lop
	import numpy as np
	
	nw=order;
	nj1=njs[0];
	nj2=njs[1];
	[n1,n2]=din.shape;

	if twoplane==0:
		dip1=np.float32(dip).flatten(order='F');
		dip2=dip1;
	else:
		dip1=np.float32(dip[:,:,0]).flatten(order='F');
		dip2=np.float32(dip[:,:,1]).flatten(order='F');
	mask=np.float32(mask).flatten(order='F');
	din=np.float32(din).flatten(order='F');
	
	dout=csoint2d(din,mask,dip1,dip2,n1,n2,nw,nj1,nj2,niter,drift,hasmask,twoplane,prec,verb);
	dout=dout.reshape(n1,n2,order='F');


	
	
	return dout






import numpy as np
def pwspray3d(din,dipi,dipx,ns2,ns3,order,eps):
	'''
	pwspray3d: 3D plane-wave spray operator 
	
	by Yangkang Chen, 2022
	
	INPUT
	din: input
	dipi: inline slope
	dipx: xline slope
	ns2/3: smoothing radius
	order: PWD order
	eps: regularization (default:0.01);
	
	OUTPUT
	dout: result
	
	'''
	[n1,n2,n3]=din.shape;
	n23=n2*n3;

	np2=2*ns2+1;	#spray diameter
	np3=2*ns3+1;	#spray diameter
	nnp=np2*np3;

	trace=np.zeros(n1);
	e=eps*eps;
	nw=order;

	# traveltime table
	t=np.zeros([np2,np3]);
	for i3 in range(0,np3):
		for i2 in range(0,np2):
			t[i2,i3] = np.hypot(i2-ns2,i3-ns3); 	#square root of the sum of squares

	visit=np.argsort(t.flatten(order='F'));
	u=np.zeros([n1,nnp,n23]);
	din=din.reshape(n1,n23,order='F')
	p1=dipi.reshape(n1,n23,order='F')
	p2=dipx.reshape(n1,n23,order='F')

	for ii in range(0,n23):
	
		u[:,ns3*np2+ns2,ii]=din[:,ii]; 	#central trace in the predicted dimension
	
		i2=np.mod(ii,n2);
		i3=np.int64(np.floor(ii/n2));

		for ip in range(0,nnp):

			[up, jp, up2, up3 ] = get_update(ip, np2, np3, t.flatten(order='F'), visit);
# 			print(ii,i2,i3,ip,up,jp,up2,up3)
			# from jp to j
			k2=np.mod(jp,np2);
			k3=np.int64(np.floor(jp/np2));
		
			j2=i2+k2-ns2;
			j3=i3+k3-ns3;
				
			if j2<0 or j2>=n2 or j3<0 or j3>=n3:
				continue;
		
			j=j2+j3*n2;
			if up & 1:
				if up2:
					if j2==0:
						continue;
					j2=j-1;
					q2=p1[:,j2];
					k2=jp-1;
				else:
					if j2==n2-1:
						continue;

					j2=j+1;
					q2=p1[:,j]; ###### error?
					k2=jp+1;
		
			if up & 2:
				if up3:
					if j3==0:
						continue;

					j3=j-n2;
					q3=p2[:,j3];
					k3=jp-np2;
				else:
					if j3==n3-1:
						continue;

					j3=j+n2;
					q3=p2[:,j];###### error?
					k3=jp+np2;

			#corresponding to the eqs.24/25
			#prediction forward or backward
			if up==0:
					continue;
			elif up==1:
					u[:,jp,j] = predict1_step(e,nw,up2,q2,u[:,k2,j2]);
			elif up==2:
					u[:,jp,j] = predict1_step(e,nw,up3,q3,u[:,k3,j3]);
			elif up==3:
					u[:,jp,j] = predict2_step(e,nw,up2,up3,q2,q3,u[:,k2,j2],u[:,k3,j3]);
		
	dout=u.reshape(n1,nnp,n2,n3,order='F');

	return dout

def get_update(ii, n1,n2, t, visit):
	'''
	forward or backward options
	
	INPUT
	ii: loop number
	n1:   [n1,n2]=size(t)
	n2:   [n1,n2]=size(t)
	t:	traveltime
	visit: sorted sequence of t
	
	OUTPUT
	jj: local linear index
	up: prediction type
	up1: forward or backward
	up2: forward or backward
	'''


	n12=n1*n2;
	jj=visit[ii];
	t1=t[jj];

	i1=np.mod(jj,n1);
	i2=np.floor(jj/n1);

	up=0;
	if n1>1:
		a1=jj-1;
		b1=jj+1;
		
		up1= (i1 and ((i1==n1-1) or (1!= (t[a1]>t[b1]))));
		if up1:
			c1=a1;
		else:
			c1=b1;

		if t1>t[c1]:
			up = up | 1; 	#Bit-wise OR.

	if n2>1:
		a2=jj-n1;
		b2=jj+n1;
	
		up2= (i2 and (i2==n2-1 or 1!= (t[a2]>t[b2])));
		if up2:
			c2=a2;
		else:
			c2=b2;

		if t1>t[c2]:
			up = up | 2; 	#Bit-wise OR.

	return up, jj, up1, up2

def predict1_step(e,nw,forw,pp,trace1):
	'''
	predict1_step: prediction step from one trace
	
	INPUT
	e: regularization parameter (default, 0.01*0.01);
	nw: accuracy order
	forw: forward or backward
	pp: slope
	trace1: input trace
	
	OUTPUT
	trace: output trace
	
	'''
	n1=trace1.shape[0];
	nb=2*nw;
	eps=e;
	eps2=e;

	diag=np.zeros(n1);
	offd=np.zeros([n1,nb]);

	diag,offd = regularization(diag,offd,nw,eps,eps2);

	w,diag,offd = pwd_define(forw,diag,offd,n1,nw,pp);


	t0=trace1[0];
	t1=trace1[1];
	t2=trace1[n1-2];
	t3=trace1[n1-1];

	trace = pwd_set(0,w,diag,offd,pp,trace1);

	trace[0]=trace[0]+eps2*t0;
	trace[1]=trace[1]+eps2*t1;

	trace[n1-2]=trace[n1-2]+eps2*t2;
	trace[n1-1]=trace[n1-1]+eps2*t3;


	trace = banded_solve(n1,nb,diag,offd,trace);   

	return trace


def regularization(diag,offd,nw,eps,eps2):
	'''
	fill diag and offd using regularization 
	
	INPUT
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	eps: regularization parameter (default: e*e, e=0.01);
	eps2: second regularization parameter (default, same as eps)
	nw: accuracy order (nb=2*nw)
	
	OUTPUT
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	'''

	nb=2*nw;
	n1=diag.size;

	for i1 in range(0,n1):
		diag[i1]=6*eps;
		offd[i1,0]=-4*eps;
		offd[i1,1]=eps;
	
		for ib in range(2,nb):
			offd[i1,ib]=0.0;

		diag[0]=eps2+eps;
		diag[1]=eps2+5*eps;
	
		diag[n1-1]=eps2+eps;
		diag[n1-2]=eps2+5*eps;  
	
		offd[0,0]=-2*eps;
		offd[n1-2,0]=-2*eps;

	return diag,offd

def pwd_define(forw,diag,offd,n1,nw,pp):
	'''
	pwd_define: matrix multiplication
	
	INPUT
	forw:forward or backward
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	n1: trace length
	nw: PWD filter(accuracy) order (default nw=1)
	pp: slope				  (1D array)
	
	OUTPUT
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	w: PWD object(struct)
	
	'''
	
	#define PWD object(struct)
	w={'n':n1,'na':2*nw+1,'a':np.zeros([n1,2*nw+1]),'b':np.zeros(2*nw+1)}
	
	n=w['n'];
	nb=2*nw;

	for i in range(0,n):
		w['b']=passfilter(pp[i],nw)[0];
		for j in range(0,w['na']):
			if forw:
				w['a'][i,j]=w['b'][w['na']-j-1];
			else:
				w['a'][i,j]=w['b'][j];

	for i in range(0,n):
		for j in range(0,w['na']):
			k=i+j-nw;
			if k>=nw and k<n-nw:
				aj=w['a'][k,j];
				diag[i]=diag[i]+aj*aj;
		for m in range(0,2*nw):
			for j in range(m+1,w['na']):
				k=i+j-nw;
				if k>=nw and k<n-nw:
					aj=w['a'][k,j];
					am=w['a'][k,j-m-1];
					offd[i,m]=offd[i,m]+am*aj;
	return w,diag,offd


def passfilter(p,nw):
	'''
	passfilter: find filter coefficients 
	All-pass plane-wave destruction filter coefficients
	
	INPUT
	p: slope
	nw: PWD filter(accuracy) order
	
	OUTPUT
	a: output filter (n+1) (1D array)
	b: temp variable of a
	
	'''

	n=nw*2;
	b=np.zeros(n+1);
	a=np.zeros(n+1);
	for k in range(0,n+1):
		bk=1;
		for j in range(0,n):
			if j<n-k:
				bk=bk*(k+j+1.0)/(2*(2*j+1)*(j+1));
			else:
				bk=bk*1.0/(2*(2*j+1));
		b[k]=bk;
	for k in range(0,n+1):
		ak=b[k];
		for j in range(0,n):
			if j<n-k:
				ak=ak*(n-j-p);
			else:
				ak=ak*(p+j+1);
		a[k]=ak;
	return a,b


def pwd_set(adj,w,diag,offd,pp,inp):
	'''
	pwd_set: matrix multiplication
	
	INPUT
	adj:adjoint flag
	w: PWD object(struct)
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	pp: slope				  (1D array)
	inp: model
	
	OUTPUT
	out: data 
	
	'''
	n=w['n'];
	nw=np.int((w['na']-1)/2);

	#	# pwd_set
	tmp=np.zeros(n);
	out=np.zeros(n);
	if adj:
		for i in range(0,n):
			tmp[i]=0.0;

		for i in range(0,n):
			for j in range(0,w['na']):
				k=i+j-nw;
				if k>=nw and k<n-nw:
					tmp[k]=tmp[k]+w['a'][k,j]*inp[i];
		for i in range(0,n):
			out[i]=0.0;
		for i in range(nw,n-nw):
			for j in range(0,w['na']):
				k=i+j-nw;
				out[k]=out[k]+w['a'][i,j]*tmp[i];
	else:
		for i in range(0,n):
			tmp[i]=0.0;

		for i in range(nw,n-nw):
			for j in range(0,w['na']):
				k=i+j-nw;
				tmp[i]=tmp[i]+w['a'][i,j]*inp[k];
		for i in range(0,n):		
			out[i]=0.0;
			for j in range(0,w['na']):
				k=i+j-nw;
				if k>=nw and k<n-nw:
					out[i]=out[i]+w['a'][k,j]*tmp[k];
	return out


def banded_solve(n,band,diag,offd,b):
	'''
	banded_solve: Banded matrix solver
	
	INPUT
	n:	matrix size
	band: band size
	diag: defined diagonal .   (1D array)
	offd: defined off-diagonal (2D array)
	b: input trace
	
	OUTPUT
	b: trace solution
	
	'''
	#define Band object(struct)
	slv={'n':n,'band':band,'d':np.zeros(n),'o':np.zeros([n-1,band])}

	# define the banded matrix
	for k in range(0,slv['n']):
		t=diag[k];
		m1=np.min([k,slv['band']]);
		for m in range(0,m1):
			t=t-slv['o'][k-m-1,m]*slv['o'][k-m-1,m]*slv['d'][k-m-1];

		slv['d'][k]=t;
		n1=np.min([slv['n']-k-1,slv['band']]);
		for n in range(0,n1):
			t=offd[k,n];
			m1=np.min([k,slv['band']-n-1]);
			for m in range(0,m1):
				t=t-slv['o'][k-m-1,m]*slv['o'][k-m-1,n+m+1]*slv['d'][k-m-1];

			slv['o'][k,n]=t/slv['d'][k];


	# the solver 
	for k in range(1,slv['n']):
		t=b[k];
		m1=np.min([k,slv['band']]);
		for m in range(0,m1):
			t=t-slv['o'][k-m-1,m]*b[k-m-1];
		b[k]=t;

	for k in range(slv['n']-1,-1,-1):
		t=b[k]/slv['d'][k];
		m1=np.min([slv['n']-k-1,slv['band']]);
		for m in range(0,m1):
			t=t-slv['o'][k,m]*b[k+m+1];

		b[k]=t;


	return b


def predict2_step(e,nw,forw1,forw2,pp1,pp2,trace1,trace2):
	'''
	predict2_step: prediction step from two trace
	
	INPUT
	e: regularization parameter (default, 0.01*0.01);
	nw: accuracy order
	forw1: forward or backward
	forw2: forward or backward
	pp1: slope
	pp2: slope
	trace1: input trace
	trace2: input trace
	
	OUTPUT
	trace: output trace
	
	'''
	n1=trace1.shape[0];
	nb=2*nw;
	eps=e;
	eps2=e;

	diag=np.zeros(n1);
	offd=np.zeros([n1,nb]);
	diag,off = regularization(diag,offd,nw,eps,eps2);

	w1,diag,offd = pwd_define(forw1,diag,offd,n1,nw,pp1);
	w2,diag,offd = pwd_define(forw2,diag,offd,n1,nw,pp2);

	t0=0.5*(trace1[0]+trace2[0]);
	t1=0.5*(trace1[1]+trace2[1]);
	t2=0.5*(trace1[n1-2]+trace2[n1-2]);
	t3=0.5*(trace1[n1-1]+trace2[n1-1]);

	tmp1 = pwd_set(0,w1,diag,offd,pp1,trace1);
	tmp2 = pwd_set(0,w2,diag,offd,pp2,trace2);

	trace=tmp1+tmp2;

	trace[0]=trace[0]+eps2*t0;
	trace[1]=trace[1]+eps2*t1;
	trace[n1-2]=trace[n1-2]+eps2*t2;
	trace[n1-1]=trace[n1-1]+eps2*t3;

	trace = banded_solve(n1,nb,diag,offd,trace);   

	return trace


import numpy as np
from bpcfun import *

def bandpass(din,dt,flo=0,fhi=0.5,nplo=6,nphi=6,phase=0,verb=1):
	'''
	bandpass: Bandpass filtering.
	
	Aug, 05, 2020
	by Yangkang Chen
	
	INPUT
	din:	  input data
	dt:       sampling
	flo:      Low frequency in band, default is 0
	fhi:      High frequency in band, default is Nyquist
	nplo=6:   number of poles for low cutoff
	nphi=6:   number of poles for high cutoff
	phase=0:  y: minimum phase, n: zero phase
	verb=0:   verbosity flag
	
	OUTPUT
	dout:     output data
	
	RFERENCE
	November 2012 program of the month: http://ahay.org/blog/2012/11/03/program-of-the-month-sfbandpass/
	
	EXAMPLE 1
	demos/test_pyseistr_das.py
	
	EXAMPLE 2
	from pyseistr import ricker;
	wav,tw=ricker(20,0.004,2);
	from pyseistr import bandpass;
	wav2=bandpass(wav,0.004,flo=0,fhi=20);
	
	import matplotlib.pyplot as plt;
	plt.subplot(2,1,1);
	plt.plot(tw,wav);plt.ylabel('Amplitude');plt.gca().set_ylim([-0.5,1]);
	plt.subplot(2,1,2);
	plt.plot(tw,wav2);plt.xlabel('Time (s)');plt.ylabel('Amplitude');plt.gca().set_ylim([-0.5,1]);
	plt.show();

	
	'''
	if din.ndim==1:	#for 1D problems
		din=np.expand_dims(din, axis=1)
		
	if din.ndim==2:	#for 2D problems
		din=np.expand_dims(din, axis=2)
		
	if din.shape[0]==1 and din.shape[1]>1 and din.shape[2]==1: #row vector
		din=din.flatten();
	
	[n1,n22,n33]=din.shape;
	
	n2=n22*n33;
	dout=np.zeros([n1,n2]);
	
	d1=dt;
	eps=0.0001;

	if flo<0:
		print('Negative flo');
	else:
		flo=flo*d1;


	if fhi<0:
		print('Negative flo');
	else:
		fhi=fhi*d1;
		if flo>fhi:
			print('Need flo < fhi\n');

		if 0.5<fhi:
			print('Need fhi < Nyquist\n');


	if nplo<1:
		nplo=1;

	if nplo>1 and not phase:
		nplo=nplo/2;

	if nphi<1:
		nphi=1;

	if nphi>1 and not phase:
		nphi=nphi/2;


	if verb:
		print("flo=%g fhi=%g nplo=%d nphi=%d\n"%(flo,fhi,nplo,nphi));
	
	if flo>eps:
		blo=butter_init(0,flo,nplo);
	else:
		blo='NULL';

	if fhi<0.5-eps:
		bhi=butter_init(1,fhi,nphi);
	else:
		bhi='NULL';

	for i2 in range(0,n2):
		trace=din[:,i2].copy(); #.copy is necessary
		if blo != 'NULL':
			trace=butter_apply(blo,n1,trace);
			if not phase:
				trace=reverse (n1, trace);
				trace=butter_apply (blo, n1, trace);
				trace=reverse (n1, trace);
	
		if bhi != 'NULL':
			trace=butter_apply(bhi,n1,trace);
			if not phase:
				trace=reverse (n1, trace);
				trace=butter_apply (bhi, n1, trace);
				trace=reverse (n1, trace);

		dout[:,i2]=trace[:,0];

	dout=dout.reshape(n1,n22,n33);

	dout=np.squeeze(dout);
	return dout

def bandpassc(din,dt,flo=0,fhi=0.5,nplo=6,nphi=6,phase=0,verb=1):
	'''
	bandpass: Bandpass filtering.
	
	Aug, 05, 2020
	by Yangkang Chen
	
	INPUT
	din:	  input data
	dt:       sampling
	flo:      Low frequency in band, default is 0
	fhi:      High frequency in band, default is Nyquist
	nplo=6:   number of poles for low cutoff
	nphi=6:   number of poles for high cutoff
	phase=0:  y: minimum phase, n: zero phase
	verb=0:   verbosity flag
	
	OUTPUT
	dout:     output data
	
	RFERENCE
	November 2012 program of the month: http://ahay.org/blog/2012/11/03/program-of-the-month-sfbandpass/
	
	EXAMPLE 1
	demos/test_pyseistr_das.py
	
	EXAMPLE 2
	from pyseistr import ricker;
	wav,tw=ricker(20,0.004,2);
	from pyseistr import bandpassc;
	wav2=bandpassc(wav,0.004,flo=0,fhi=20);
	
	import matplotlib.pyplot as plt;
	figsize=(8, 8)
	plt.subplot(2,1,1);
	plt.plot(tw,wav);plt.ylabel('Amplitude');plt.gca().set_ylim([-0.5,1]);
	plt.subplot(2,1,2);
	plt.plot(tw,wav2);plt.xlabel('Time (s)');plt.ylabel('Amplitude');plt.gca().set_ylim([-0.5,1]);
	plt.show();
	
	'''
	if din.ndim==1:	#for 1D problems
		din=np.expand_dims(din, axis=1)
		
	if din.ndim==2:	#for 2D problems
		din=np.expand_dims(din, axis=2)
		
	if din.shape[0]==1 and din.shape[1]>1 and din.shape[2]==1: #row vector
		din=din.flatten();

	[n1,n2,n3]=din.shape;
		
	din=np.float32(din).flatten(order='F');
	dout=cbp(din,n1,n2,n3,dt,flo,fhi,nplo,nphi,phase,verb);
	dout=dout.reshape(n1,n2,n3,order='F');

	if n3==1:
		dout=np.squeeze(dout);
	return dout
	
def butter_init(low,cutoff,nn):
	'''
	butter_init: initialize
	Aug, 5, 2020
	Yangkang Chen
	
	INPUT
	low:	  low-pass (or high-pass)
	cutoff:   cut off frequency
	nn:	   number of poles
	
	OUTPUT
	bw:	   butterworth struct
	
	'''

	arg=2*np.pi*cutoff;
	sinw=np.sin(arg);
	cosw=np.cos(arg);

	bw={'nn':nn,'low':low,'den':np.zeros([2,int(np.floor((nn+1)/2))])};
	
	if np.mod(nn,2)>0:
		if low:
			fact=(1+cosw)/sinw;
			bw['den'][0,int(np.floor(nn/2))]=1./(1.+fact);
			bw['den'][1,int(np.floor(nn/2))]=1.-fact;
		else:
			fact=sinw/(1.+cosw);
			bw['den'][0,int(np.floor(nn/2))]=1./(fact+1.0);
			bw['den'][1,int(np.floor(nn/2))]=fact-1.0;


	fact=ifnot(low,np.sin(0.5*arg),np.cos(0.5*arg));
	fact=fact*fact;

	for j in range(0,int(np.floor(nn/2))):
		ss=np.sin(np.pi*(2*j+1)/(2*nn))*sinw;
		bw['den'][0,j]=fact/(1.+ss);
		bw['den'][1,j]=(1-ss)/fact;

	bw['mid']=-2.*cosw/fact;
	return bw


def butter_apply(bw,nx,x):
	'''
	butter_apply: filter the data (in place)
	
	Implementation is inspired by D. Hale and J.F. Claerbout, 1983, Butterworth
	dip filters: Geophysics, 48, 1033-1038.
	
	Aug, 5, 2020
	Yangkang Chen
	
	INPUT
	bw: butterworth struct
	nx: size of x
	x: input data
	
	OUTPUT
	x: output data
	'''
	d1=bw['mid'];
	nn=bw['nn'];
	if np.mod(nn,2)>0:
		d0=bw['den'][0,int(np.floor(nn/2))];
		d2=bw['den'][1,int(np.floor(nn/2))];
		x0=0;
		y1=0;
		for ix in range(0,nx):
			x1=x0;x0=x[ix].copy(); #This bug is striking, .copy is necessary because otherwise x0 will automatically change according to y0.
			y0=ifnot(bw['low'],(x0 + x1 - d2 * y1)*d0,(x0 - x1 - d2 * y1)*d0);
			x[ix]=y0;
			y1=y0;
	for j in range(0,int(np.floor(nn/2))):
		d0=bw['den'][0,j];
		d2=bw['den'][1,j];
		x1=0;x0=0;y1=0;y2=0;
		for ix in range(0,nx):
			x2=x1;x1=x0;x0=x[ix].copy();#This bug is striking, .copy is necessary because otherwise x0 will automatically change according to y0.
			y0=ifnot(bw['low'],(x0 + 2*x1 + x2 - d1 * y1 - d2 * y2)*d0,(x0 - 2*x1 + x2 - d1 * y1 - d2 * y2)*d0);
			y2=y1;x[ix]=y0;y1=y0;
	return x

def reverse(n1,trace):
	'''
	reverse a trace (in place)
	'''
	for i1 in range(0,int(np.floor(n1/2))):
		t=trace[i1].copy();#This bug is striking, .copy is necessary because otherwise x0 will automatically change according to y0.
		trace[i1]=trace[n1-i1-1];
		trace[n1-i1-1]=t;
	return trace


def ifnot(yes, v1, v2):
	'''
	ifnot: equivalent to C grammar (v=yes?v1:v2)
	Yangkang Chen
	July, 22, 2020
	'''
	if yes:
		v=v1;
	else:
		v=v2;

	return v


import numpy as np
from dipcfun import *

def dip2d(din,niter=5,liter=20,order=2,eps_dv=0.01, eps_cg=1, tol_cg=0.000001,rect=[10,10,1],verb=1):
	'''
	dip2d: dip estimation based on shaping regularized PWD algorithm 
	(independent implementation)
	
	INPUT
	din: input data (nt*nx)
	niter: number of nonlinear iterations
	liter: number of linear iterations (in divn)
	order: accuracy order
	eps_dv: eps for divn  (default: 0.01)
	eps_cg: eps for CG    (default: 1)
	tol_cg: tolerence for CG (default: 0.000001)
	rect:  smoothing radius (ndim*1)
	verb: verbosity flag
	
	OUTPUT
	dip:  2D slope
	
	NOTE: mask function is not enabled yet (will do in the future)
	'''
	
	from .divne import divne
	
	p0 = 0.0;
	nj1 = 1;
	nj2 = 1;
	dim = 3;
	
	n = np.zeros(dim,dtype='int');
	n1 = din.shape[0];
	n2 = din.shape[1];
	n[0] = n1;
	n[1] = n2;
	n[2] = 1;

	n123 = din.size;

	ratio=np.zeros([n1,n2,1]);
	dip=np.zeros([n1,n2,1]);

	for iter in range(0,niter):
		"""corresponding to the eq.21 in the paper  (Wang et al., 2022)"""
		[u1,u2] = conv_allpass(din,dip,order); 
		ratio = divne(-u2, u1, liter, rect, n, eps_dv, eps_cg, tol_cg,verb);
		dip=dip+ratio;

	return dip

def dip2dc(din,niter=5,liter=20,order=2,eps_dv=0.01, eps_cg=1, tol_cg=0.000001,rect=[10,10,1],verb=1,mask=None):
	'''
	dip2dc: dip estimation based on shaping regularized PWD algorithm 
	(C implementation)
	
	INPUT
	din: input data (nt*nx)
	niter: number of nonlinear iterations
	liter: number of linear iterations (in divn)
	order: accuracy order
	eps_dv: eps for divn  (default: 0.01)
	eps_cg: eps for CG    (default: 1)
	tol_cg: tolerence for CG (default: 0.000001)
	rect:  smoothing radius (ndim*1)
	verb: verbosity flag
	
	OUTPUT
	dip:  2D slope
	
	from .divne import divne
	'''
	
	p0 = 0.0;
	nj1 = 1;
	nj2 = 1;
	dim = 3;
	
	n1 = din.shape[0];
	n2 = din.shape[1];
	n3 = 1;
	
	r1=rect[0]
	r2=rect[1]
	r3=rect[2]

	if mask is None:
		din=np.float32(din.flatten(order='F'));
		hasmask=0
	else:
		if din.size != mask.size:
			Exception("Mask and Data should have the same dimension")
		else:
			din=np.float32(np.concatenate([din.flatten(order='F'),mask.flatten(order='F')],axis=0));
			hasmask=1;
	dip=dipc(din,n1,n2,n3,niter,liter,order,eps_dv,eps_cg,tol_cg,r1,r2,r3,hasmask,verb);
	dip=dip.reshape(n1,n2,order='F');
		
	return dip
	
def conv_allpass(din,dip,order):
	'''
	conv_allpass: Convolutional operator implemented by an allpass filter
	
	Linearized inverse problem
	C'(\sigma)d\Delta \sigma =  C(\sigma)d
	
	IPNUT:
	din: input data
	dip: 3D dip
	order: accuracy order
	
	OUTPUT:
	u1: C'(\sigma)d (denominator)
	u2: C(\sigma)d  (numerator)
	
	'''
	[n1,n2]=din.shape;
	u1=np.zeros([n1,n2]);
	u2=np.zeros([n1,n2]);

	if order==1:
		nw=1;
	else:
		nw=2;

	filt1=np.zeros(2*nw+1);
	filt2=np.zeros(2*nw+1);

	#cprresponding to eqs.17-19
	#destruct the data
	for i1 in range(nw,n1-nw):
		for i2 in range(0,n2-1):
			if order==1:
				filt1=B3d(dip[i1,i2]);
				filt2=B3(dip[i1,i2]);
			else:
				filt1=B5d(dip[i1,i2]);
				filt2=B5(dip[i1,i2]);

			for iw in range(-nw,nw+1):
				u1[i1,i2]=u1[i1,i2]+(din[i1+iw,i2+1]-din[i1-iw,i2])*filt1[iw+nw];
				u2[i1,i2]=u2[i1,i2]+(din[i1+iw,i2+1]-din[i1-iw,i2])*filt2[iw+nw];

	return u1,u2

#cprresponding to eqs.13-16
#form the filters
def B3(sigma):
	'''
	B3 coefficient
	sigma: slope
	'''
	b3=np.zeros[2];
	b3[0]=(1-sigma)*(2-sigma)/12;
	b3[1]=(2+sigma)*(2-sigma)/6;
	b3[2]=(1+sigma)*(2+sigma)/12;
	
	return b3

def B3d(sigma):
	'''
	B3 coefficient derivative
	sigma: slope
	'''
	b3d=np.zeros[2];
	b3d[0]=-(2-sigma)/12-(1-sigma)/12;
	b3d[1]=(2-sigma)/6-(2+sigma)/6;
	b3d[2]=(2+sigma)/12+(1+sigma)/12;

	return b3d

def B5(sigma):
	'''
	B5 coefficient
	sigma: slope
	'''
	b5=np.zeros(5);
	b5[0]=(1-sigma)*(2-sigma)*(3-sigma)*(4-sigma)/1680;
	b5[1]=(4-sigma)*(2-sigma)*(3-sigma)*(4+sigma)/420;
	b5[2]=(4-sigma)*(3-sigma)*(3+sigma)*(4+sigma)/280;
	b5[3]=(4-sigma)*(2+sigma)*(3+sigma)*(4+sigma)/420;
	b5[4]=(1+sigma)*(2+sigma)*(3+sigma)*(4+sigma)/1680;

	return b5

def B5d(sigma):
	'''
	B5 coefficient derivative
	sigma: slope
	'''
	b5d=np.zeros(5);
	b5d[0]=-(2-sigma)*(3-sigma)*(4-sigma)/1680-\
	(1-sigma)*(3-sigma)*(4-sigma)/1680-\
	(1-sigma)*(2-sigma)*(4-sigma)/1680-\
	(1-sigma)*(2-sigma)*(3-sigma)/1680;

	b5d[1]=-(2-sigma)*(3-sigma)*(4+sigma)/420-\
	(4-sigma)*(3-sigma)*(4+sigma)/420-\
	(4-sigma)*(2-sigma)*(4+sigma)/420+\
	(4-sigma)*(2-sigma)*(3-sigma)/420;

	b5d[2]=-(3-sigma)*(3+sigma)*(4+sigma)/280-\
	(4-sigma)*(3+sigma)*(4+sigma)/280+\
	(4-sigma)*(3-sigma)*(4+sigma)/280+\
	(4-sigma)*(3-sigma)*(3+sigma)/280;

	b5d[3]=-(2+sigma)*(3+sigma)*(4+sigma)/420+\
	(4-sigma)*(3+sigma)*(4+sigma)/420+\
	(4-sigma)*(2+sigma)*(4+sigma)/420+\
	(4-sigma)*(2+sigma)*(3+sigma)/420;

	b5d[4]=(2+sigma)*(3+sigma)*(4+sigma)/1680+\
	(1+sigma)*(3+sigma)*(4+sigma)/1680+\
	(1+sigma)*(2+sigma)*(4+sigma)/1680+\
	(1+sigma)*(2+sigma)*(3+sigma)/1680;

	return b5d



from sofcfun import *

def somean2d(dn,dip,ns,order,eps):
	'''
	somean2d: plane-wave smoothing 
	
	INPUT
	dn: model   noisy data
	dip: slope (2D array)
	ns:       spray radius
	order:    PWD order
	eps: regularization (default:0.01);

	OUTPUT
	ds:  smoothed data
	
	EXAMPLE
	
	
	'''
	import numpy as np
	n1=dn.shape[0];
	n2=dn.shape[1];

	ns2=2*ns+1;	#spray diameter

	#flattening
	from .pwspray2d import pwspray2d
	
	utmp=pwspray2d(dn,dip,ns,order,eps);
	u=utmp.reshape(n1,ns2,n2,order='F')
	ds=np.sum(u,1)/ns2;

	return ds

def somean2dc(dn,dip,ns,order,eps,adj=0,verb=1):
	'''
	somean2dc: plane-wave smoothing implemented in C
	
	INPUT
	dn: model   noisy data
	dip: slope (2D array)
	ns:       spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	adj: adjoint flat (default: 0, forward)
	verb: verbosity
	
	OUTPUT
	ds:  smoothed data
	
	EXAMPLE
	
	'''
	import numpy as np

	ns2=2*ns+1;	#spray diameter
	
	if dn.ndim==2:
		[n1,n2]=dn.shape;
		n3=1;
	else: #assuming ndim=3;
		[n1,n2,n3]=dn.shape;
	
	dn=np.float32(dn).flatten(order='F');
	dip=np.float32(dip).flatten(order='F');
	
	ds=csomean2d(dn,dip,n1,n2,n3,ns,order,adj,eps,verb);
	ds=ds.reshape(n1,n2,n3,order='F')
	
	if n3==1:	#for 2D problems
		ds=np.squeeze(ds)
	return ds
	
	
	


from sofcfun import *

def somf2d(dn,dip,ns,order,eps,option=1):
	'''
	somf2d: structure-oriented median filter
	
	INPUT:
	dn: model   noisy data
	dip: slope (2D array)
	ns:       spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	option=1 or 2: (1 for MF; 2 for SVMF)
	
	OUTPUT
	ds:  filtered data
	 
	REFERENCES
	Huang et al., 2021, Erratic noise suppression using iterative structure-oriented space-varying median filtering with sparsity constraint, Geophysical Prospecting, 69, 101-121.
	Chen et al., 2020, Deblending of simultaneous-source data using a structure-oriented space-varying median filter, Geophysical Journal International, 222, 1805–1823.
	Gan et al., 2016, Separation of simultaneous sources using a structural-oriented median filter in the flattened dimension, Computers & Geosciences, 86, 46-54.
	Chen, Y., 2015, Deblending using a space-varying median filter, Exploration Geophysics, 46, 332-341.
	
	EXAMPLE
	demos/test_xxx_somf2d.py
	'''
	
	import numpy as np
	from .pwspray2d import pwspray2d
	from .mf import mf
	from .mf import svmf
	
	n1=dn.shape[0];
	n2=dn.shape[1];
	ns2=2*ns+1;	#spray diameter

	#flattening
	utmp=pwspray2d(dn,dip,ns,order,eps);
	u=utmp.reshape(n1,ns2,n2,order='F')


	for i2 in range(0,n2):
		if option==1:
			u[:,:,i2]=mf(u[:,:,i2],ns2,1,2);
		else:
			u[:,:,i2],win_tmp=svmf(u[:,:,i2],ns2,1,2);
	
	ds=u[:,ns,:];
	
	return ds

def somf2dc(dn,dip,ns,order,eps,option=1,verb=1):
	'''
	somf2dc: structure-oriented median filter in C
	
	INPUT
	dn: model   noisy data
	dip: slope (2D array)
	ns:       spray radius
	order:    PWD order
	eps: regularization (default:0.01);
	option=1 or 2: (1 for MF; 2 for SVMF)
	
	OUTPUT
	ds:  filtered data
	 
	REFERENCES
	Huang et al., 2021, Erratic noise suppression using iterative structure-oriented space-varying median filtering with sparsity constraint, Geophysical Prospecting, 69, 101-121.
	Chen et al., 2020, Deblending of simultaneous-source data using a structure-oriented space-varying median filter, Geophysical Journal International, 222, 1805–1823.
	Gan et al., 2016, Separation of simultaneous sources using a structural-oriented median filter in the flattened dimension, Computers & Geosciences, 86, 46-54.
	Chen, Y., 2015, Deblending using a space-varying median filter, Exploration Geophysics, 46, 332-341.
	
	EXAMPLE
	demos/test_pyseistr_somf2d.py
	'''
	
	import numpy as np
	ns2=2*ns+1;	#spray diameter
	
	if dn.ndim==2:
		[n1,n2]=dn.shape;
		n3=1;
	else: #assuming ndim=3;
		[n1,n2,n3]=dn.shape;
	
	dn=np.float32(dn).flatten(order='F');
	dip=np.float32(dip).flatten(order='F');
	
	ds=csomf2d(dn,dip,n1,n2,n3,ns,2*ns+1,option,order,eps,verb);
	ds=ds.reshape(n1,n2,n3,order='F')
	
	if n3==1:	#for 2D problems
		ds=np.squeeze(ds)
	return ds
	

	
	
	


import numpy as np
def solver(opL,solv,nx,ny,x,dat,niter,par_L,par):
	'''
	Generic linear solver. (same as Madagascar function: sf_solver)
	
	Solves
	L{x}    =~ dat
	
	Yangkang Chen
	Aug, 05, 2020
	Ported to Python in Apr, 25, 2022
	
	INPUT
	opL: forward/linear operator
	solv: stepping function
	nx:   size of x   (1D vector)
	ny:   size of dat (1D vector)
	x:    estimated model
	dat:  data
	niter:number of iterations
	
	par. (parameter struct)
	  "wt":     float*:         weight      
	  "wght":   sf_weight wght: weighting function
	  "x0":     float*:         initial model
	  "nloper": sf_operator:    nonlinear operator
	  "mwt":    float*:         model weight
	  "verb":   bool:           verbosity flag
	  "known":  bool*:          known model mask
	  "nmem":   int:            iteration memory
	  "nfreq":  int:            periodic restart
	  "xmov":   float**:        model iteration
	  "rmov":   float**:        residual iteration
	  "err":    float*:         final error
	  "res":    float*:         final residual
	
	OUPUT
	x: estimated model
	
	'''
	
	TOLERANCE=1.e-12;
	forget=0;
	x=x.flatten(order='F');
	dat=dat.flatten(order='F');
	xmov=None;
	rmov=None;
	wht=None;
	err=None;
	res=None;

	if 'wt' in par:
		wt=par['wt'];
	else:
		wt=None;
	
	if 'wght' in par:
		wght=par['wght'];
	else:
		wght=None;
	
	if 'x0' in par:
		x0=par['x0'];
	else:
		x0=None;
	
	if 'nloper' in par:
		nloper=par['nloper'];
		par_nloper=par['par_nloper'];
	else:
		nloper=None;
		
	if 'mwt' in par:
		mwt=par['mwt'];
	else:
		mwt=None;
		
	if 'verb' in par:
		verb=par['verb'];
	else:
		verb=0;
		
	if 'known' in par:
		known=par['known'];
	else:
		known=None;

	if 'nmem' in par:
		nmem=par['nmem'];
	else:
		nmem=-1;
	
	if 'nfreq' in par:
		nfreq=par['nfreq'];
	else:
		nfreq=0;
		
	if 'xmov' in par:
		xmov=par['xmov'];
	else:
		xmov=None;

	if 'rmov' in par:
		rmov=par['rmov'];
	else:
		rmov=None;
		
	if 'err' in par:
		err=par['err'];
	else:
		err=None;

	if 'res' in par:
		res=par['res'];
	else:
		res=None;
		
	g=np.zeros(nx);
	rr=np.zeros(ny);
	gg=np.zeros(ny);
	
	if wt  is not None or wght  is not None:
		td=zeros(ny);
		if wt  is not None:
			wht=wt;
		else:
			wht=np.ones(ny);
	
	if mwt  is not None:
		g2=np.zeros(nx);
	
	rr=-dat;
	if x0  is not None:
		x=x0;
		if mwt  is not None:
			x=x*mwt;
		if nloper  is not None:
			par_nloper['d']=rr;
			rr=nloper(x,par_nloper,0,1);
		else:
			par_L['d']=rr;
			rr=opL(x,par_L,0,1);
	else:
		x=np.zeros(nx);
	
	dpr0=np.sum(rr*rr);
	dpg0=1.0;
	
	for n in range(0,niter):
		if nmem>=0:
			forget= (n >= nmem);
			
		if wght  is not None and forget:
			wht=wght(ny,rr); #wght is a function
		
		if wht  is not None:
			rr=rr*wht;
			td=rr*wht;
			g=opL(td,par_L,1,0);
		else:
			g=opL(rr,par_L,1,0);
		
		if mwt  is not None:
			g=g*mwt;
			
		if known  is not None:
			for ii in range(0,nx):
				if known[ii]:
					g[ii]=0.0;
		
		if mwt  is not None:
			g2=g*mwt;
			gg=opL(g2,par_L,0,0);
		else:
			gg=opL(g,par_L,0,0);
		
		if wht  is not None:
			gg=gg*wht;
		
		if forget and (nfreq !=0): #periodic restart
			forget = (0 == np.mod(n+1,nfreq));
		
		if n==0:
			dpg0=np.sum(g*g);
			dpr=1.0;
			dpg=1.0;
		else:
			dpr=np.sum(rr*rr)/dpr0;
			dpg=np.sum(g*g)/dpg0;
		
		if verb:
			print('iteration %d res %f mod %f grad %f !'%(n+1, dpr,np.sum(x*x), dpg));
			
		if dpr < TOLERANCE or dpg < TOLERANCE:
			if verb:
				print('convergence in %d iterations\n',n+1);
			if mwt  is not None:
				x=x*mwt;
			break;
		
		x,rr = solv(forget,nx,ny,x,g,rr,gg);
		
		forget=0;
		if nloper  is not None:
			rr=-dat;
			if mwt  is not None:
				x=x*mwt;
			par_nloper['d']=rr;
			rr=nloper(x,par_nloper,0,1); 
		else:
			if wht is not None:
				rr=-dat;
				if mwt  is not None:
					x=x*mwt;
				par_L['d']=rr;
				rr=opL(x,par_L,0,1);
			else:
				if mwt  is not None and ( xmov  is not None or n==niter-1 ):
					x=x*mwt;
	
		if xmov  is not None:
			xmov[:,n]=x;
			par['xmov']=xmov;
		
		if rmov  is not None:
			rmov[:,n]=rr;
			par['rmov']=rmov;
		
		if err  is not None:
			err[n]=np.sum(r*r);
			par['err']=err;

	if res  is not None:
		res=rr;
		par['res']=res;
		
	return x,par

def conjgrad(opP,opL,opS, p, x, dat, eps_cg, tol_cg, N,ifhasp0,par_P,par_L,par_S,verb):
	'''
	conjgrad: conjugate gradient with shaping
	
	by Yangkang Chen, 2022
	
	Modified by Yangkang Chen, Nov, 09, 2019 (fix the "adding" for each oper)
	
	INPUT
	opP: preconditioning operator
	opL: forward/linear operator
	opS: shaping operator
	d:  data
	N:  number of iterations
	eps_cg:  scaling
	tol_cg:  tolerance
	ifhasp0: flag indicating if has initial model
	par_P: parameters for P
	par_L: parameters for L
	par_S: parameters for S
	verb: verbosity flag
	
	OUPUT
	x: estimated model
	
	'''

	nnp=p.size;
	nx=par_L['nm'];	#model size
	nd=par_L['nd'];	#data size

	if opP  is not None:
		d=-dat; #nd*1
		r=opP(d,par_P,0,0);
	else:
		r=-dat;  

	if ifhasp0:
		x=op_S(p,par_S,0,0);
		if opP  is not None:
			d=opL(x,par_L,0,0);
			par_P['d']=r;#initialize data
			r=opP(d,par_P,0,1);
		else:
			par_P['d']=r;#initialize data
			r=opL(x,par_L,0,1);

	else:
		p=np.zeros(nnp);#define np!
		x=np.zeros(nx);#define nx!

	dg=0;
	g0=0;
	gnp=0;
	r0=np.sum(r*r);   #nr*1

	for n in range(1,N+1):
		gp=eps_cg*p; #np*1
		gx=-eps_cg*x; #nx*1
		
		if opP is not None:
			d=opP(r,par_P,1,0);#adjoint
			par_L['m']=gx;#initialize model
			gx=opL(d,par_L,1,1);#adjoint,adding
		else:
			par_L['m']=gx;#initialize model
			gx=opL(r,par_L,1,1);#adjoint,adding

	
		par_S['m']=gp;#initialize model
		gp=opS(gx,par_S,1,1);#adjoint,adding
		gx=opS(gp.copy(),par_S,0,0);#forward,adding
		#The above gp.copy() instead of gp is the most striking bug that has been found because otherwise gp was modified during the shaping operation (opS) (Mar, 28, 2022)
		
		if opP is not None:
			d=opL(gx,par_P,0,0);#forward
			gr=opP(d,par_L,0,0);#forward
		else:
			gr=opL(gx,par_L,0,0);#forward

		gn = np.sum(gp*gp); #np*1

		if n==1:
			g0=gn;
			sp=gp; #np*1
			sx=gx; #nx*1
			sr=gr; #nr*1
		else:
			alpha=gn/gnp;
			dg=gn/g0;
		
			if alpha < tol_cg or dg < tol_cg:
				return x;
				break;
		
			gp=alpha*sp+gp;
			t=sp;sp=gp;gp=t;
		
			gx=alpha*sx+gx;
			t=sx;sx=gx;gx=t;
		
			gr=alpha*sr+gr;
			t=sr;sr=gr;gr=t;

	 
		beta=np.sum(sr*sr)+eps_cg*(np.sum(sp*sp)-np.sum(sx*sx));
		
		if verb:
			print('iteration: %d, res: %g !'%(n,np.sum(r* r) / r0));  

		alpha=-gn/beta;
	
		p=alpha*sp+p;
		x=alpha*sx+x;
		r=alpha*sr+r;
	
		gnp=gn;

	return x


def cgstep(forget,nx,ny,x,g,rr,gg):
	'''
	Step of conjugate-gradient iteration.
	Yangkang Chen
	Aug, 05, 2020
	
	INPUT
	forget:restart flag
	nx:   model size
	ny:   data size
	x:    current model [nx]
	g:    gradient [nx]
	rr:   data residual [ny]
	gg:   conjugate gradient [ny]
	
	OUTPUT
	x:    current model [nx]
	g:    gradient [nx]
	rr:   data residual [ny]
	gg:   conjugate gradient [ny]
	'''
	EPSILON=1.e-12;
	Allocated=0;
	if not Allocated:
		Allocated =1;
		forget=1;
		S=np.zeros(nx);
		Ss=np.zeros(ny);
	if forget:
		S=np.zeros(nx);
		Ss=np.zeros(ny);
		beta=0;
		alfa=np.sum(gg*gg);
		#Solve G . ( R + G*alfa) = 0
		if alfa<=0:
			return x,rr
		alfa=-np.sum(gg*rr)/alfa;
	else:
		#search plane by solving 2-by-2
		#G . (R + G*alfa + S*beta) = 0
		#S . (R + G*alfa + S*beta) = 0
		gdg=np.sum(gg*gg);
		sds=np.sum(Ss*Ss);
		gds=np.sum(gg*Ss);
		if gdg == 0 or sds==0:
			return x,rr
			
		determ=1.0-(gds/gdg)*(gds/sds);
		if determ > EPSILON:
			determ = determ * gdg * sds;
		else:
			determ = gdg * sds * EPSILON;
		gdr=-np.sum(gg*rr);
		sdr=-np.sum(Ss*rr);
		
		alfa = ( sds * gdr - gds * sdr ) / determ;
		beta = (-gds * gdr + gdg * sdr ) / determ;
	S=S*beta;
	S=alfa*g+S;
	Ss=Ss*beta;
	Ss=alfa*gg+Ss;
	x=x+S;
	rr=rr+Ss;
	return x,rr


import numpy as np
def fkdip(d,w):
	'''
	fkdip: FK dip filter
	
	INPUT
	d: 	input data (2D)
	w:  half width (in percentage) of the cone filter (i.e., w*nk=nwidth)
	
	OUTPUT
	d0: output data
	
	EXAMPLE
	import pyseistr as ps
	from pyseistr import gensyn
	from pyseistr import genflat
	import numpy as np
	data=gensyn();[nt,nx]=data.shape;
	noise=genflat(nt,nx,t=np.linspace(5,10*36,37,dtype='int32'),amp=0.5*np.ones(37),freq=80);
	datan=data+noise;
	datafk=datan-ps.fkdip(datan,0.005); ## FK filtering
	import matplotlib.pyplot as plt;	
	plt.figure(figsize=(8, 8));
	plt.subplot(1,3,1);
	plt.imshow(datan,aspect='auto');plt.xlabel('Trace');plt.ylabel('Time sample');
	plt.subplot(1,3,2);
	plt.imshow(datafk,aspect='auto');plt.xlabel('Trace');
	plt.subplot(1,3,3);
	plt.imshow(datan-datafk,aspect='auto');plt.xlabel('Trace');
	plt.show();
	
	REFERENCE
	Chen, Y., Savvaidis, A., Fomel, S., Chen, Y., Saad, O.M., Wang, H., Oboué, Y.A.S.I., Yang, L. and Chen, W., 2023. Denoising of Distributed Acoustic Sensing Seismic Data Using an Integrated Framework. Seismological Society of America, 94(1), pp.457-472.
	
	'''
	
	[n1,n2]=d.shape;
	nf=nextpow2(n1);
	nk=nextpow2(n2);
	nf2=int(nf/2);
	nk2=int(nk/2);
	Dfft1=np.fft.fft(d,nf,0);
	Dtmp=Dfft1[0:nf2+1,:];

	#processing area
	Dtmp2=np.fft.fft(Dtmp,nk,1);
	Dtmp2=np.fft.fftshift(Dtmp2,1);

	nw=w*nk;
	[nn1,nn2]=Dtmp2.shape;
	mask=np.zeros([nn1,nn2]);

	for i1 in range(1,nn1+1):
		for i2 in range(1,nn2+1):
			if i1> (nn1/nw)*(i2-(nk2)) and i1> (nn1/nw)*((nk2)-i2):
				mask[i1-1,i2-1]=1;

	Dtmp2=Dtmp2*mask;
	Dtmp=np.fft.ifft(np.fft.ifftshift(Dtmp2,1),nk,1);

	#honor symmetry for inverse fft
	Dfft2=np.zeros([nf,nk],dtype=np.complex_);
	Dfft2[0:nf2+1,:]=Dtmp;
	Dfft2[nf2+1:,:]=np.conj(np.flipud(Dtmp[1:-1,:]));
	d0=np.real(np.fft.ifft(Dfft2,nf,0));
	d0=d0[0:n1,0:n2];

	return d0
	
def nextpow2(N):
    """ Function for finding the next power of 2 """
    n = 1
    while n < N: n *= 2
    return n
    


#!/usr/bin/env python
# -*- encoding: utf8 -*-
import io
import os

from setuptools import setup
from distutils.core import Extension
import numpy


long_description = """
Source code: https://github.com/aaspip/pyseistr""".strip() 


def read(*names, **kwargs):
    return io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")).read()

dipc_module = Extension('dipcfun', sources=['pyseistr/src/dip_cfuns.c'], 
										include_dirs=[numpy.get_include()])
sofc_module = Extension('sofcfun', sources=['pyseistr/src/sof_cfuns.c'], 
										include_dirs=[numpy.get_include()])
sofc3d_module = Extension('sof3dcfun', sources=['pyseistr/src/sof3d_cfuns.c'], 
										include_dirs=[numpy.get_include()])
sointc3d_module = Extension('soint3dcfun', sources=['pyseistr/src/soint3d_cfuns.c'], 
										include_dirs=[numpy.get_include()])
sointc2d_module = Extension('soint2dcfun', sources=['pyseistr/src/soint2d_cfuns.c'], 
										include_dirs=[numpy.get_include()])
bpc_module = Extension('bpcfun', sources=['pyseistr/src/bp_cfuns.c'], 
										include_dirs=[numpy.get_include()])
cohc_module = Extension('cohcfun', sources=['pyseistr/src/coh_cfuns.c'], 
										include_dirs=[numpy.get_include()])
setup(
    name="pyseistr",
    version="0.0.4.4.0",
    license='GNU General Public License, Version 3 (GPLv3)',
    description="A python package for structural denoising and interpolation of multi-channel seismic data",
    long_description=long_description,
    author="pyseistr developing team",
    author_email="chenyk2016@gmail.com",
    url="https://github.com/aaspip/pyseistr",
    ext_modules=[dipc_module,sofc_module,sofc3d_module,sointc2d_module,sointc3d_module,bpc_module,cohc_module],
    packages=['pyseistr'],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    keywords=[
        "seismology", "earthquake seismology", "exploration seismology", "array seismology", "denoising", "science", "engineering", "structure", "local slope", "filtering"
    ],
    install_requires=[
        "numpy", "scipy", "matplotlib"
    ],
    extras_require={
        "docs": ["sphinx", "ipython", "runipy"]
    }
)



import numpy as np
def allpass3_lop(din,par,adj,add):
	#3-D Plane-wave destruction filter
	if adj==1:
		d=din;
		if 'm' in par and add==1:
			m=par['m'];
		else:
			m=np.zeros(par['nm']);
	else:
		m=din;
		if 'd' in par and add==1:
			d=par['d'];
		else:
			d=np.zeros(par['nd']);

	nm=par['nm'];	 #int
	nd=par['nd'];	 #int
	ap1=par['ap1'];  #int
	ap2=par['ap2'];  #int
	n1=nm;

	[ xx,yy ] = adjnull( adj,add,nm,nd,m,d );

	nx = ap1['nx'];
	ny = ap1['ny'];
	nz = ap1['nz'];
	nw = ap1['nw'];
	nj = ap1['nj'];

	for iz in range(0,ap1['nz']):
		for iy in range(0,ap1['ny']-1):
			for ix in range(ap1['nw']*ap1['nj'],ap1['nx']-ap1['nw']*ap1['nj']):
				ii=ix+ap1['nx']*(iy+ap1['ny']*iz);
			
				if ap1['drift']:
					id=SF_NINT(ap1['pp'][ii]);
					if ix-ap1['nw']*ap1['nj']-id<0 or ix+ap1['nw']*ap1['nj']-id>=nx:
						continue;
					ap1['flt'],tmp=passfilter(ap1['pp'][ii]-id, ap1['nw'] );
				
					for iw in range(0,2*ap1['nw']+1):
						iss = (iw-ap1['nw'])*ap1['nj'];
					
						if adj:
							xx[ii+iss+nx] = xx[ii+iss+nx]+yy[ii]*ap1['flt'][iw];
							xx[ii-iss-id] = xx[ii-iss-id]-yy[ii]*ap1['flt'][iw];
						else:
							yy[ii] = yy[ii] + (xx[ii+iss+nx] - xx[ii-iss-id])*ap1['flt'][iw];
				else:
					ap1['flt'],tmp=passfilter(ap1['pp'][ii],ap1['nw']);
				
					for iw in range(0,2*ap1['nw']+1):
						iss = (iw-ap1['nw'])*ap1['nj'];
						if adj:
							xx[ii+iss+nx] = xx[ii+iss+nx] + yy[ii]*ap1['flt'][iw];
							xx[ii-iss]	= xx[ii-iss]	- yy[ii]*ap1['flt'][iw];
						else:
							yy[ii] = yy[ii]+(xx[ii+iss+nx] - xx[ii-iss])*ap1['flt'][iw];
							
	nx = ap2['nx'];
	ny = ap2['ny'];
	nz = ap2['nz'];
	nw = ap2['nw'];
	nj = ap2['nj'];

	if nx*ny*nz !=n1:
		print('size mismatch');


	for iz in range(0,ap2['nz']-1):
		for iy in range(0,ap2['ny']):
			for ix in range(ap2['nw']*ap2['nj'],ap2['nx']-ap2['nw']*ap2['nj']):
				ii=ix+ap2['nx']*(iy+ap2['ny']*iz);
			
				if ap2['drift']:
					id=SF_NINT(ap2['pp'][ii]);
					if ix-ap2['nw']*ap2['nj']-id<0 or ix+ap2['nw']*ap2['nj']-id>=nx:
						continue;

					ap2['flt'],tmp=passfilter(ap2['pp'][ii]-id, ap2['nw'] );
				
					for iw in range(0,2*ap2['nw']+1):
						iss = (iw-ap2['nw'])*ap2['nj'];
					
						if adj:
							xx[ii+iss+nx*ny] = xx[ii+iss+nx]+yy[ii+n1]*ap2['flt'][iw];
							xx[ii-iss-id] = xx[ii-iss-id]-yy[ii+n1]*ap2['flt'][iw];
						else:
							yy[ii+n1] = yy[ii+n1] + (xx[ii+iss+nx*ny] - xx[ii-iss-id])*ap2['flt'][iw];

				else:
					ap2['flt'],tmp=passfilter(ap2['pp'][ii],ap2['nw']);
				
					for iw in range(0,2*ap2['nw']+1):
						iss = (iw-ap2['nw'])*ap2['nj'];
						if adj:
							xx[ii+iss+nx*ny] = xx[ii+iss+nx*ny] + yy[ii+n1]*ap2['flt'][iw];
							xx[ii-iss]	= xx[ii-iss]	- yy[ii+n1]*ap2['flt'][iw];
						else:
							yy[ii+n1] = yy[ii+n1]+(xx[ii+iss+nx*ny] - xx[ii-iss])*ap2['flt'][iw];



	if adj==1:
		dout=xx;
	else:
		dout=yy;

	return dout


def passfilter(p,nw):
	'''
	passfilter: find filter coefficients£®verfied)
	All-pass plane-wave destruction filter coefficients
	
	INPUT
	p: slope
	nw: filter order
	
	OUTPUT
	a: output filter (n+1) (1D array)
	
	'''
	n=nw*2;
	b=np.zeros([n+1,1]);
	a=np.zeros([n+1,1]);
	
	for k in range(0,n+1):
		bk=1;
		for j in range(0,n):
			if (j<n-k):
				bk=bk*(k+j+1.0)/(2*(2*j+1)*(j+1));
			else:
				bk=bk*1.0/(2*(2*j+1));

		b[k]=bk;

	for k in range(0,n+1):
		ak=b[k];
		for j in range(0,n):
			if j<n-k:
				ak=ak*(n-j-p);
			else:
				ak=ak*(p+j+1);
		a[k]=ak;

	return a,b


def adjnull( adj,add,nm,nd,m,d ):
	'''
	Claerbout-style adjoint zeroing Zeros out the output (unless add is true). 
	Useful first step for and linear operator.
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) and later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	adj : adjoint flag; add: addition flag; nm: size of m; nd: size of d
	'''
	if add:
		return m,d

	if adj:
		m=np.zeros(nm);
		for i in range(0,nm):
			m[i] = 0.0;
	else:
		d=np.zeros(nd);
		for i in range(0,nd):
			d[i] = 0.0;

	return m,d



import numpy as np
def divne(num, den, Niter, rect, ndat, eps_dv, eps_cg, tol_cg,verb):
	'''
	divne: N-dimensional smooth division rat=num/den		  
	This is a subroutine from the seistr package (https://github.com/chenyk1990/seistr)
	
	by Yangkang Chen, 2022, verified to be exactly the same as the Matlab version
	
	INPUT
	num: numerator
	den: denominator
	Niter: number of iterations
	rect: triangle radius [ndim], e.g., [5,5,1]
	ndat: data dimensions [ndim], e.g., [n1,n2,1]
	eps_dv: eps for divn  (default: 0.01)
	eps_cg: eps for CG	(default: 1)
	tol_cg: tolerence for CG (default: 0.000001)
	verb: verbosity flag
	
	OUTPUT
	rat: output ratio
	 
	REFERENCE
	H. Wang, Y. Chen, O. Saad, W. Chen, Y. Oboue, L. Yang, S. Fomel, and Y. Chen, 2022, A Matlab code package for 2D/3D local slope estimation and structural filtering. Geophysics, 87(3), F1–F14, doi: 10.1190/geo2021-0266.1. 
	'''
	n=num.size
	
	ifhasp0=0
	p=np.zeros(n)
	
	num=num.reshape(n,order='F')
	den=den.reshape(n,order='F')
	
	if eps_dv > 0.0:
		for i in range(0,n):
			norm = 1.0 / np.hypot(den[i], eps_dv);
			num[i] = num[i] * norm;
			den[i] = den[i] * norm;
	norm=np.sum(den*den);
	if norm == 0.0:
		rat=np.zeros(n);
		return rat
	norm = np.sqrt(n / norm);
	num=num*norm;
	den=den*norm;
		
	par_L={'nm':n,'nd':n,'w':den}
	par_S={'nm':n,'nd':n,'nbox':rect,'ndat':ndat,'ndim':3}
	
	from .solvers import conjgrad
	rat = conjgrad(None, weight_lop, trianglen_lop, p, None, num, eps_cg, tol_cg, Niter,ifhasp0,[],par_L,par_S,verb);
	rat=rat.reshape(ndat[0],ndat[1],ndat[2],order='F')

	return rat


def weight_lop(din,par,adj,add):
	'''
	weight_lop: Weighting operator (verified)
	
	Ported to Python by Yangkang Chen, 2022
	
	INPUT
	din: model/data
	par: parameter
	adj: adj flag
	add: add flag
	
	OUTPUT
	dout: data/model
	'''
	nm=par['nm'];
	nd=par['nd'];
	w=par['w'];

	if adj==1:
		d=din;
		if 'm' in par and add==1:
			m=par['m'];
		else:
			m=np.zeros(par['nm']);
	else:
		m=din;
		if 'd' in par and add==1:
			d=par['d'];
		else:
			d=np.zeros(par['nd']);
	m,d  = adjnull( adj,add,nm,nd,m,d );
	if adj==1:
		m=m+d*w; #dot product
	else: #forward
		d=d+m*w; #d becomes model, m becomes data


	if adj==1:
		dout=m;
	else:
		dout=d;

	return dout
	
def trianglen_lop(din,par,adj,add ):
	'''
	trianglen_lop: N-D triangle smoothing operator (verified)
	
	Ported to Python by Yangkang Chen, 2022
	
	INPUT
	din: model/data
	par: parameter
	adj: adj flag
	add: add flag
	
	OUTPUT
	dout: data/model
	'''
	if adj==1:
		d=din;
		if 'm' in par and add==1:
			m=par['m'];
		else:
			m=np.zeros(par['nm']);
	else:
		m=din;
		if 'd' in par and add==1:
			d=par['d'];
		else:
			d=np.zeros(par['nd']);


	nm=par['nm'];	 #int
	nd=par['nd'];	 #int
	ndim=par['ndim']; #int
	nbox=par['nbox']; #vector[ndim]
	ndat=par['ndat']; #vector[ndim]

	[ m,d ] = adjnull( adj,add,nm,nd,m,d );

	tr = [];

	s =[1,ndat[0],ndat[0]*ndat[1]];

	for i in range(0,ndim):
		if (nbox[i] > 1):
			nnp = ndat[i] + 2*nbox[i];
			wt = 1.0 / (nbox[i]*nbox[i]);
			tr.append({'nx':ndat[i], 'nb':nbox[i], 'box':0, 'np':nnp, 'wt':wt, 'tmp':np.zeros(nnp)});
		else:
			tr.append('NULL');

	if adj==1:
		tmp=d;
	else:
		tmp=m;

	for i in range(0,ndim):
		if tr[i] != 'NULL':
			for j in range(0,int(nd/ndat[i])):
				i0=first_index(i,j,ndim,ndat,s);
				[tmp,tr[i]]=smooth2(tr[i],i0,s[i],0,tmp);
	
	if adj==1:
		m=m+tmp;
	else:
		d=d+tmp;
		
	if adj==1:
		dout=m;
	else:
		dout=d;

	return dout


def first_index( i, j, dim, n, s ):
	'''
	first_index: Find first index for multidimensional transforms
	Ported to Python by Yangkang Chen, 2022
	
	INPUT
	i:	dimension [0...dim-1]
	j:	line coordinate
	dim:  number of dimensions
	n:	box size [dim], vector
	s:	step [dim], vector
	
	OUTPUT
	i0:   first index
	'''

	n123 = 1;
	i0 = 0;
	for k in range(0,dim):
		if (k == i):
			continue;
		ii = np.floor(np.mod((j/n123), n[k]));
		n123 = n123 * n[k];
		i0 = i0 + ii * s[k];

	return int(i0)


def smooth2( tr, o, d, der, x):
	'''
	smooth2: apply triangle smoothing
	
	Ported to Python by Yangkang Chen, 2022
	
	INPUT
	tr:   smoothing object
	o:	trace sampling
	d:	trace sampling
	x:	data (smoothed in place)
	der:  if derivative
	
	OUTPUT
	x: smoothed result
	tr: triangle struct
	'''

	tr['tmp'] = triple2(o, d, tr['nx'], tr['nb'], x, tr['tmp'], tr['box'], tr['wt']);
	tr['tmp'] = doubint2(tr['np'], tr['tmp'], (tr['box'] or der));
	x = fold2(o, d, tr['nx'], tr['nb'], tr['np'], x, tr['tmp']);

	return x,tr


def triple2( o, d, nx, nb, x, tmp, box, wt ):
	'''
	BY Yangkang Chen, Nov, 04, 2019
	'''
	for i in range(0,nx+2*nb):
		tmp[i] = 0;

	if box:
		tmp[1:]	 = cblas_saxpy(nx,  +wt,x[o:],d,tmp[1:],   1); 	#y += a*x
		tmp[2*nb:]  = cblas_saxpy(nx,  -wt,x[o:],d,tmp[2*nb:],1);
	else:
		tmp		 = cblas_saxpy(nx,  -wt,x[o:],d,tmp,	   1); 	#y += a*x
		tmp[nb:]	= cblas_saxpy(nx,2.*wt,x[o:],d,tmp[nb:],  1);
		tmp[2*nb:]  = cblas_saxpy(nx,  -wt,x[o:],d,tmp[2*nb:],1);

	return tmp

def doubint2( nx, xx, der ):
	'''
	Modified by Yangkang Chen, Nov, 04, 2019
	'''
	#integrate forward
	t = 0.0;
	for i in range(0,nx):
		t = t + xx[i];
		xx[i] = t;

	if der:
		return xx

	#integrate backward
	t = 0.0;
	for i in range(nx-1,-1,-1):
		t = t + xx[i];
		xx[i] = t

	return xx



def cblas_saxpy( n, a, x, sx, y, sy ):
	'''
	y += a*x
	Modified by Yangkang Chen, Nov, 04, 2019
	'''
	for i in range(0,n):
		ix = i * sx;
		iy = i * sy;
		y[iy] = y[iy] + a * x[ix];

	return y

def fold2(o, d, nx, nb, np, x, tmp):
	'''
	Modified by Yangkang Chen, Nov, 04, 2019
	'''
	#copy middle
	for i in range(0,nx):
		x[o+i*d] = tmp[i+nb];

	#reflections from the right side
	for j in range(nb+nx,np+1,nx):
		if (nx <= np-j):
			for i in range(0,nx):
				x[o+(nx-1-i)*d] = x[o+(nx-1-i)*d] + tmp[j+i];
		else:
			for i in range(0,np-j):
				x[o+(nx-1-i)*d] = x[o+(nx-1-i)*d] + tmp[j+i];
		j = j + nx;
		if (nx <= np-j):
			for i in range(0,nx): 
				x[o+i*d] = x[o+i*d] + tmp[j+i];
		else:
			for i in range(0,np-j):
				x[o+i*d] = x[o+i*d] + tmp[j+i];

	#reflections from the left side
	for j in range(nb,-1,-nx):
		if (nx <= j):
			for i in range(0,nx):
				x[o+i*d] = x[o+i*d] + tmp[j-1-i];
		else:
			for i in range(0,j):
				x[o+i*d] = x[o+i*d] + tmp[j-1-i];
		j = j - nx;
		if (nx <= j):
			for i in range(0,nx):
				x[o+(nx-1-i)*d] = x[o+(nx-1-i)*d] + tmp[j-1-i];
		else:
			for i in range(0,j):
				x[o+(nx-1-i)*d] = x[o+(nx-1-i)*d] + tmp[j-1-i];
	return x

def adjnull( adj,add,nm,nd,m,d ):
	'''
	Claerbout-style adjoint zeroing Zeros out the output (unless add is true). 
	Useful first step for and linear operator.
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) and later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	adj : adjoint flag; add: addition flag; nm: size of m; nd: size of d
	'''
	
	if add:
		return m,d

	if adj:
		m=np.zeros(nm);
		for i in range(0,nm):
			m[i] = 0.0;
	else:
		d=np.zeros(nd);
		for i in range(0,nd):
			d[i] = 0.0;

	return m,d




	
